from PyQt5.QtWidgets import (
    QButtonGroup,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QLabel,
    QRadioButton,
    QVBoxLayout,
    QWidget,
)

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import BooleanParameter, RealParameter

from ...model.ptychi import (
    PtyChiAutodiffSettings,
    PtyChiDMSettings,
    PtyChiDeviceRepository,
    PtyChiEnumerators,
    PtyChiLSQMLSettings,
    PtyChiSettings,
)
from ..parametric import (
    CheckBoxParameterViewController,
    CheckableGroupBoxParameterViewController,
    ComboBoxParameterViewController,
    DecimalLineEditParameterViewController,
    DecimalSliderParameterViewController,
    ParameterViewController,
    SpinBoxParameterViewController,
)

__all__ = ['PtyChiReconstructorViewController']


class PtyChiDeviceViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        use_devices: BooleanParameter,
        repository: PtyChiDeviceRepository,
        *,
        tool_tip: str = '',
    ) -> None:
        super().__init__(use_devices, 'Use Devices', tool_tip=tool_tip)
        layout = QVBoxLayout()

        for device in repository:
            device_label = QLabel(device)
            layout.addWidget(device_label)

        self.get_widget().setLayout(layout)


class PtyChiPrecisionParameterViewController(ParameterViewController, Observer):
    def __init__(self, use_double_precision: BooleanParameter, *, tool_tip: str = '') -> None:
        super().__init__()
        self._use_double_precision = use_double_precision
        self._single_precision_button = QRadioButton('Single')
        self._double_precision_button = QRadioButton('Double')
        self._button_group = QButtonGroup()
        self._widget = QWidget()

        self._single_precision_button.setToolTip('Compute using single precision')
        self._double_precision_button.setToolTip('Compute using double precision')

        if tool_tip:
            self._widget.setToolTip(tool_tip)

        self._button_group.addButton(self._single_precision_button, 1)
        self._button_group.addButton(self._double_precision_button, 2)
        self._button_group.setExclusive(True)
        self._button_group.idToggled.connect(self._sync_view_to_model)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self._single_precision_button)
        layout.addWidget(self._double_precision_button)
        layout.addStretch()
        self._widget.setLayout(layout)

        self._sync_model_to_view()
        use_double_precision.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_view_to_model(self, tool_id: int, checked: bool) -> None:
        if tool_id == 2:
            self._use_double_precision.set_value(checked)

    def _sync_model_to_view(self) -> None:
        button = self._button_group.button(2 if self._use_double_precision.get_value() else 1)

        if button is not None:
            button.setChecked(True)

    def _update(self, observable: Observable) -> None:
        if observable is self._use_double_precision:
            self._sync_model_to_view()


class PtyChiMomentumAccelerationGradientMixingFactorViewController(
    CheckableGroupBoxParameterViewController
):
    def __init__(
        self,
        use_gradient_mixing_factor: BooleanParameter,
        gradient_mixing_factor: RealParameter,
        *,
        tool_tip: str = '',
    ) -> None:
        super().__init__(
            use_gradient_mixing_factor,
            'Use Gradient Mixing Factor',
            tool_tip='Controls how the current gradient is mixed with the accumulated velocity in LSQML momentum acceleration',
        )
        self._gradient_mixing_factor_view_controller = DecimalLineEditParameterViewController(
            gradient_mixing_factor, tool_tip=tool_tip
        )

        layout = QVBoxLayout()
        layout.addWidget(self._gradient_mixing_factor_view_controller.get_widget())
        self.get_widget().setLayout(layout)


class PtyChiReconstructorViewController(ParameterViewController):
    def __init__(
        self,
        settings: PtyChiSettings,
        autodiff_settings: PtyChiAutodiffSettings | None,
        dm_settings: PtyChiDMSettings | None,
        lsqml_settings: PtyChiLSQMLSettings | None,
        enumerators: PtyChiEnumerators,
        repository: PtyChiDeviceRepository,
    ) -> None:
        super().__init__()
        self._num_epochs_view_controller = SpinBoxParameterViewController(
            settings.num_epochs, tool_tip='Number of epochs to run'
        )
        self._num_sync_epochs_view_controller = SpinBoxParameterViewController(
            settings.num_sync_epochs, tool_tip='Number of epochs between updates'
        )
        self._batch_size_view_controller = SpinBoxParameterViewController(
            settings.batch_size, tool_tip='Number of data to process in each minibatch'
        )
        self._batching_mode_view_controller = ComboBoxParameterViewController(
            settings.batching_mode, enumerators.batching_modes(), tool_tip='Batching mode to use'
        )
        self._compact_mode_update_clustering_view_controller = SpinBoxParameterViewController(
            settings.compact_mode_update_clustering,
            tool_tip='When greater than zero, the number of epochs between updating clusters in compact batching mode',
        )
        self._device_view_controller = PtyChiDeviceViewController(
            settings.use_devices, repository, tool_tip='Default device to use for computation'
        )
        self._compute_precision_view_controller = PtyChiPrecisionParameterViewController(
            settings.use_double_precision,
            tool_tip='Floating point precision to use for computation',
        )
        self._fft_precision_view_controller = PtyChiPrecisionParameterViewController(
            settings.use_double_precision_for_fft,
            tool_tip='Floating point precision to use for critical FFT operations',
        )
        self.allow_nondeterministic_algorithms_view_controller = CheckBoxParameterViewController(
            settings.allow_nondeterministic_algorithms,
            'Allow Nondeterministic Algorithms',
            tool_tip='When checked, nondeterministic algorithms will be used. This may lead to different results on different runs',
        )

        self._use_low_memory_view_controller = CheckBoxParameterViewController(
            settings.use_low_memory_mode,
            'Use Low Memory Mode',
            tool_tip='When checked, forward propagation of ptychography will be done using less vectorized code. This reduces the speed, but also lowers memory usage',
        )
        self._pad_for_shift_view_controller = SpinBoxParameterViewController(
            settings.pad_for_shift,
            tool_tip='Number of pixels to pad arrays (with border values) before shifting',
        )

        self._use_far_field_propagation_view_controller = CheckBoxParameterViewController(
            settings.use_far_field_propagation,
            'Use Far Field Propagation',
            tool_tip='When checked, far field propagation will be used instead of near field propagation',
        )
        self._fft_shift_diffraction_patterns_view_controller = CheckBoxParameterViewController(
            settings.fft_shift_diffraction_patterns,
            'FFT Shift Diffraction Patterns',
            tool_tip='When checked, the diffraction patterns will be FFT-shifted',
        )
        self._save_data_on_device_view_controller = CheckBoxParameterViewController(
            settings.save_data_on_device,
            'Save Data on Device',
            tool_tip='When checked, diffraction data will be saved on the device',
        )
        self._widget = QGroupBox('Reconstructor')

        layout = QFormLayout()
        layout.addRow('Number of Epochs:', self._num_epochs_view_controller.get_widget())
        layout.addRow('Update Interval:', self._num_sync_epochs_view_controller.get_widget())

        if dm_settings is None:
            layout.addRow('Batch Size:', self._batch_size_view_controller.get_widget())
            layout.addRow('Batch Mode:', self._batching_mode_view_controller.get_widget())
            layout.addRow(
                'Update Clustering:',
                self._compact_mode_update_clustering_view_controller.get_widget(),
            )

        layout.addRow(self._device_view_controller.get_widget())
        layout.addRow('Compute Precision:', self._compute_precision_view_controller.get_widget())
        layout.addRow('FFT Precision:', self._fft_precision_view_controller.get_widget())
        layout.addRow(self.allow_nondeterministic_algorithms_view_controller.get_widget())

        layout.addRow(self._use_low_memory_view_controller.get_widget())
        layout.addRow('Pad For Shift:', self._pad_for_shift_view_controller.get_widget())

        layout.addRow(self._use_far_field_propagation_view_controller.get_widget())
        layout.addRow(self._fft_shift_diffraction_patterns_view_controller.get_widget())
        layout.addRow(self._save_data_on_device_view_controller.get_widget())

        if autodiff_settings is not None:
            self._loss_function_view_controller = ComboBoxParameterViewController(
                autodiff_settings.loss_function,
                enumerators.loss_functions(),
                tool_tip='Loss function to optimize',
            )
            layout.addRow('Loss Function:', self._loss_function_view_controller.get_widget())

            self._forward_model_class_view_controller = ComboBoxParameterViewController(
                autodiff_settings.forward_model_class,
                enumerators.forward_models(),
                tool_tip='Forward model class',
            )
            layout.addRow('Forward Model:', self._forward_model_class_view_controller.get_widget())

        if dm_settings is not None:
            self._exit_wave_update_relaxation_view_controller = (
                DecimalSliderParameterViewController(
                    dm_settings.exit_wave_update_relaxation,
                    tool_tip='Relaxation multiplier for the exit wave update',
                )
            )
            layout.addRow(
                'Exit Wave Update Relaxation:',
                self._exit_wave_update_relaxation_view_controller.get_widget(),
            )

            self._chunk_length_view_controller = SpinBoxParameterViewController(
                dm_settings.chunk_length,
                tool_tip='Number of probe positions used in each chunk of the exit wave update loop',
            )
            layout.addRow('Chunk Length:', self._chunk_length_view_controller.get_widget())

        if lsqml_settings is not None:
            self._noise_model_view_controller = ComboBoxParameterViewController(
                lsqml_settings.noise_model,
                enumerators.noise_models(),
                tool_tip='Noise model to use',
            )
            layout.addRow('Noise Model:', self._noise_model_view_controller.get_widget())

            self._gaussian_noise_deviation_view_controller = DecimalLineEditParameterViewController(
                lsqml_settings.gaussian_noise_deviation,
                tool_tip='Standard deviation of the Gaussian noise',
            )
            layout.addRow(
                'Gaussian Noise Deviation:',
                self._gaussian_noise_deviation_view_controller.get_widget(),
            )

            self._solve_object_probe_step_size_jointly_for_first_slice_in_multislice_view_controller = CheckBoxParameterViewController(
                lsqml_settings.solve_object_probe_step_size_jointly_for_first_slice_in_multislice,
                'Solve Object Probe Step Size Jointly For First Slice In Multislice',
                tool_tip='When checked, the object and probe step length calculation will be solved simultaneously',
            )
            layout.addRow(
                self._solve_object_probe_step_size_jointly_for_first_slice_in_multislice_view_controller.get_widget()
            )

            self._solve_step_sizes_only_using_first_probe_mode_view_controller = CheckBoxParameterViewController(
                lsqml_settings.solve_step_sizes_only_using_first_probe_mode,
                'Solve Step Sizes Only Using First Probe Mode',
                tool_tip='When checked, the step sizes will be calculated using only the first probe mode',
            )
            layout.addRow(
                self._solve_step_sizes_only_using_first_probe_mode_view_controller.get_widget()
            )

            self._momentum_acceleration_gain_view_controller = (
                DecimalLineEditParameterViewController(
                    lsqml_settings.momentum_acceleration_gain,
                    tool_tip='Gain of momentum accleration',
                )
            )
            layout.addRow(
                'Momentum Acceleration Gain:',
                self._momentum_acceleration_gain_view_controller.get_widget(),
            )

            self._momentum_acceleration_gradient_mixing_factor_view_controller = PtyChiMomentumAccelerationGradientMixingFactorViewController(
                lsqml_settings.use_momentum_acceleration_gradient_mixing_factor,
                lsqml_settings.momentum_acceleration_gradient_mixing_factor,
                tool_tip='Controls how the current gradient is mixed with the accumulated velocity in LSQML momentum acceleration',
            )
            layout.addRow(
                self._momentum_acceleration_gradient_mixing_factor_view_controller.get_widget()
            )

            self._rescale_probe_intensity_in_first_epoch_view_controller = CheckBoxParameterViewController(
                lsqml_settings.rescale_probe_intensity_in_first_epoch,
                'Rescale Probe Intensity In First Epoch',
                tool_tip='When checked, the probe intensity will be rescaled on the first epoch',
            )
            layout.addRow(self._rescale_probe_intensity_in_first_epoch_view_controller.get_widget())

            self._preconditioning_damping_factor_view_controller = (
                DecimalLineEditParameterViewController(
                    lsqml_settings.preconditioning_damping_factor,
                    tool_tip='Damping factor for preconditioning the object update',
                )
            )
            layout.addRow(
                'Preconditioning Damping Factor:',
                self._preconditioning_damping_factor_view_controller.get_widget(),
            )

        self._widget.setLayout(layout)

    def get_widget(self) -> QWidget:
        return self._widget



from collections.abc import Sequence

# TODO from pvaccess import Channel, PvObjectQueue

from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition

from .builder import ProbePositionsBuilder
from .settings import ProbePositionsSettings


class StreamingScanBuilder(ProbePositionsBuilder):
    def __init__(
        self, settings: ProbePositionsSettings, point_seq: Sequence[ProbePosition]
    ) -> None:
        super().__init__(settings, 'Streaming')
        self._point_list = list(point_seq)

    def append(self, point: ProbePosition) -> None:
        self._point_list.append(point)

    def extend(self, point_seq: Sequence[ProbePosition]) -> None:
        self._point_list.extend(point_seq)

    def build(self) -> ProbePositionSequence:
        return ProbePositionSequence(self._point_list)


# TODO def echo(self, value: int = 125) -> None:
# TODO     print(f'{value=}')
# TODO
# TODO def foo(self) -> None:
# TODO     channelName = 'foo'
# TODO     ch = Channel(channelName)
# TODO     connected = ch.isConnected()
# TODO     isActive = ch.isMonitorActive()
# TODO     ch.setMonitorMaxQueueLength(3)
# TODO     ch.subscribe('echo', self.echo)
# TODO     ch.startMonitor()
# TODO     ch.stopMonitor()
# TODO     ch.unsubscribe('echo')
# TODO
# TODO     counterDict = ch.getMonitorCounters()



from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class PtychoPINNModelSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtychoPINNModel')
        self._group.add_observer(self)

        self.gridsize = self._group.create_integer_parameter('GridSize', 1, minimum=1, maximum=5)
        self.n_filters_scale = self._group.create_integer_parameter(
            'NumFiltersScale', 2, minimum=1, maximum=4
        )
        self.amp_activation = self._group.create_string_parameter('AmpActivation', 'sigmoid')
        self.object_big = self._group.create_boolean_parameter('ObjectBig', True)
        self.probe_big = self._group.create_boolean_parameter('ProbeBig', True)
        self.probe_mask = self._group.create_boolean_parameter('ProbeMask', False)
        self.pad_object = self._group.create_boolean_parameter('PadObject', True)
        self.probe_scale = self._group.create_real_parameter('ProbeScale', 4.0, minimum=0.0)
        self.gaussian_smoothing_sigma = self._group.create_real_parameter(
            'GaussianSmoothingSigma', 0.0, minimum=0.0
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtychoPINNTrainingSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtychoPINNTraining')
        self._group.add_observer(self)

        self.nphotons = self._group.create_real_parameter('NumPhotons', 1e6)  # TODO remove
        self.batch_size = self._group.create_integer_parameter(
            'BatchSize', 16, minimum=1, maximum=1 << 30
        )  # must be positive powers of two
        self.nepochs = self._group.create_integer_parameter('NumEpochs', 50, minimum=1)
        self.mae_weight = self._group.create_real_parameter(
            'MAEWeight', 0.0, minimum=0.0, maximum=1.0
        )
        self.nll_weight = self._group.create_real_parameter(
            'NLLWeight', 1.0, minimum=0.0, maximum=1.0
        )
        self.realspace_mae_weight = self._group.create_real_parameter(
            'RealspaceMAEWeight', 0.0, minimum=0.0, maximum=1.0
        )
        self.realspace_weight = self._group.create_real_parameter(
            'RealspaceWeight', 0.0, minimum=0.0, maximum=1.0
        )
        self.positions_provided = self._group.create_boolean_parameter('PositionsProvided', True)
        self.probe_trainable = self._group.create_boolean_parameter('ProbeTrainable', False)
        self.intensity_scale_trainable = self._group.create_boolean_parameter(
            'IntensityScaleTrainable', True
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtychoPINNInferenceSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtychoPINNInference')
        self._group.add_observer(self)

        self.model_path = self._group.create_path_parameter('ModelPath', Path('/path/to/model.zip'))
        self.n_nearest_neighbors = self._group.create_integer_parameter(
            'NumNearestNeighbors', 7, minimum=0
        )
        self.n_samples = self._group.create_integer_parameter('NumSamples', 1, minimum=1)

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from __future__ import annotations

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QCheckBox,
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QGroupBox,
    QHBoxLayout,
    QHeaderView,
    QLabel,
    QMenu,
    QPushButton,
    QTreeView,
    QVBoxLayout,
    QWidget,
    QWizardPage,
)


class DetectorView(QGroupBox):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__('Detector', parent)


class PatternsButtonBox(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.load_button = QPushButton('Load')
        self.load_menu = QMenu()
        self.save_button = QPushButton('Save')
        self.close_button = QPushButton('Close')
        self.analyze_button = QPushButton('Analyze')
        self.analyze_menu = QMenu()

        self.load_button.setMenu(self.load_menu)
        self.analyze_button.setMenu(self.analyze_menu)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.load_button)
        layout.addWidget(self.save_button)
        layout.addWidget(self.close_button)
        layout.addWidget(self.analyze_button)
        self.setLayout(layout)


class OpenDatasetWizardPage(QWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self._is_complete = False

    def isComplete(self) -> bool:  # noqa: N802
        """Overrides QWizardPage.isComplete()"""
        return self._is_complete

    def _set_complete(self, complete: bool) -> None:
        if self._is_complete != complete:
            self._is_complete = complete
            self.completeChanged.emit()


class OpenDatasetWizardMetadataPage(OpenDatasetWizardPage):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.detector_extent_check_box = QCheckBox('Detector Extent')
        self.detector_pixel_size_check_box = QCheckBox('Detector Pixel Size')
        self.detector_bit_depth_check_box = QCheckBox('Detector Bit Depth')
        self.detector_distance_check_box = QCheckBox('Detector Distance')
        self.pattern_crop_center_check_box = QCheckBox('Pattern Crop Center')
        self.pattern_crop_extent_check_box = QCheckBox('Pattern Crop Extent')
        self.probe_energy_check_box = QCheckBox('Probe Energy')
        self.probe_photon_count_check_box = QCheckBox('Probe Photon Count')
        self.exposure_time_check_box = QCheckBox('Exposure Time')

        self.setTitle('Import Metadata')

        layout = QVBoxLayout()
        layout.addWidget(self.detector_extent_check_box)
        layout.addWidget(self.detector_pixel_size_check_box)
        layout.addWidget(self.detector_bit_depth_check_box)
        layout.addWidget(self.detector_distance_check_box)
        layout.addWidget(self.pattern_crop_center_check_box)
        layout.addWidget(self.pattern_crop_extent_check_box)
        layout.addWidget(self.probe_energy_check_box)
        layout.addWidget(self.probe_photon_count_check_box)
        layout.addWidget(self.exposure_time_check_box)
        layout.addStretch()
        self.setLayout(layout)


class DatasetFileLayoutDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.tree_view = QTreeView()
        self.button_box = QDialogButtonBox()

        tree_header = self.tree_view.header()

        if tree_header is not None:
            tree_header.setDefaultAlignment(Qt.AlignmentFlag.AlignCenter)
            tree_header.setSectionResizeMode(QHeaderView.ResizeMode.ResizeToContents)

        self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.button_box.accepted.connect(self.accept)

        layout = QVBoxLayout()
        layout.addWidget(self.tree_view)
        layout.addWidget(self.button_box)
        self.setLayout(layout)


class SimulateDiffractionDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.product_combo_box = QComboBox()
        self.button_box = QDialogButtonBox()

        self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.button_box.accepted.connect(self.accept)
        self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.button_box.rejected.connect(self.reject)

        layout = QFormLayout()
        layout.addRow('Product:', self.product_combo_box)
        layout.addRow(self.button_box)
        self.setLayout(layout)

        self.setWindowTitle('Simulate Diffraction Patterns')


class PatternsView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.detector_view = DetectorView()
        self.tree_view = QTreeView()
        self.info_label = QLabel()
        self.button_box = PatternsButtonBox()
        self.simulate_dialog = SimulateDiffractionDialog(self)

        tree_view_header = self.tree_view.header()

        if tree_view_header is not None:
            tree_view_header.setDefaultAlignment(Qt.AlignmentFlag.AlignCenter)

        layout = QVBoxLayout()
        layout.addWidget(self.detector_view)
        layout.addWidget(self.tree_view)
        layout.addWidget(self.info_label)
        layout.addWidget(self.button_box)
        self.setLayout(layout)



from pathlib import Path
import logging

import h5py
import numpy

from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    DiffractionDataset,
    DiffractionFileReader,
    DiffractionFileWriter,
    DiffractionMetadata,
    DiffractionArray,
    DiffractionPatterns,
    DiffractionIndexes,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.tree import SimpleTreeNode

logger = logging.getLogger(__name__)


class H5DiffractionPatternArray(DiffractionArray):
    def __init__(
        self, label: str, indexes: DiffractionIndexes, file_path: Path, data_path: str
    ) -> None:
        super().__init__()
        self._label = label
        self._indexes = indexes
        self._file_path = file_path
        self._data_path = data_path

    def get_label(self) -> str:
        return self._label

    def get_indexes(self) -> DiffractionIndexes:
        return self._indexes

    def get_patterns(self) -> DiffractionPatterns:
        with h5py.File(self._file_path, 'r') as h5_file:
            try:
                item = h5_file[self._data_path]
            except KeyError:
                raise ValueError(
                    f'Failed to find diffraction pattern array at "{self._data_path}"!'
                )

            if isinstance(item, h5py.Dataset):
                logger.debug(f'Reading "{item.name}"...')
                logger.debug(f'\tcompression = {item.compression}')
                logger.debug(f'\tcompression_opts = {item.compression_opts}')

                dataset_id = item.id
                dataset_creation_properties = dataset_id.get_create_plist()
                nfilters = dataset_creation_properties.get_nfilters()
                missing_filter_names: list[str] = []

                if nfilters > 0:
                    logger.debug('\tfilters = [')

                    for filter_idx in range(nfilters):
                        filter_ = dataset_creation_properties.get_filter(filter_idx)
                        filter_code = filter_[0]
                        flags = filter_[1]
                        aux_data = filter_[2]
                        name = filter_[3].decode()
                        logger.debug(
                            f'\t\t({filter_idx}) {filter_code=}, {flags=}, {aux_data=}, {name=}'
                        )

                        if not h5py.h5z.filter_avail(filter_code):
                            missing_filter_names.append(name)

                    logger.debug('\t]')
                else:
                    logger.debug('\tfilters = []')

                if missing_filter_names:
                    names = ' '.join(missing_filter_names)
                    raise RuntimeError(f'Missing filters needed to read dataset: {names}!')

                return item[:]
            else:
                raise ValueError(f'Path {self._file_path}:{self._data_path} is not a dataset!')


class H5DiffractionFileTreeBuilder:
    def _add_attributes(
        self, tree_node: SimpleTreeNode, attribute_manager: h5py.AttributeManager
    ) -> None:
        for name, value in attribute_manager.items():
            item_details = ''

            if isinstance(value, str):
                item_details = f'STRING = "{value}"'
            elif isinstance(value, h5py.Empty):
                logger.debug(f'Skipping empty attribute {name}.')
            elif isinstance(value, numpy.ndarray):
                item_details = f'ARRAY = {value}'
            else:
                string_info = h5py.check_string_dtype(value.dtype)

                if string_info:
                    item_details = f'STRING = "{value.decode(string_info.encoding)}"'
                else:
                    item_details = f'SCALAR {value.dtype} = {value}'

            tree_node.create_child([str(name), 'Attribute', item_details])

    def create_root_node(self) -> SimpleTreeNode:
        return SimpleTreeNode.create_root(['Name', 'Type', 'Details'])

    def build(self, h5_file: h5py.File) -> SimpleTreeNode:
        root_node = self.create_root_node()
        unvisited: list[tuple[SimpleTreeNode, h5py.Group]] = [(root_node, h5_file)]

        while unvisited:
            parent_item, h5_group = unvisited.pop()

            for item_name in h5_group:
                item_type = 'Unknown'
                item_details = ''
                h5_item = h5_group.get(item_name, getlink=True)

                tree_node = parent_item.create_child([])

                if isinstance(h5_item, h5py.HardLink):
                    item_type = 'Hard Link'
                    h5_item = h5_group.get(item_name, getlink=False)

                    if isinstance(h5_item, h5py.Group):
                        item_type = 'Group'
                        self._add_attributes(tree_node, h5_item.attrs)
                        unvisited.append((tree_node, h5_item))
                    elif isinstance(h5_item, h5py.Dataset):
                        item_type = 'Dataset'
                        self._add_attributes(tree_node, h5_item.attrs)
                        space_id = h5_item.id.get_space()

                        if space_id.get_simple_extent_type() == h5py.h5s.SCALAR:
                            value = h5_item[()]

                            if isinstance(value, bytes):
                                item_details = value.decode()
                            elif isinstance(value, numpy.ndarray):
                                item_details = f'STRING = {h5_item.asstr()}'
                            else:
                                string_info = h5py.check_string_dtype(value.dtype)

                                if string_info:
                                    item_details = (
                                        f'STRING = "{value.decode(string_info.encoding)}"'
                                    )
                                else:
                                    item_details = f'SCALAR {value.dtype} = {value}'
                        elif h5_item.size == 1:
                            try:
                                value = h5_item[()]
                            except Exception:
                                item_details = f'{h5_item.shape} {h5_item.dtype}'
                            else:
                                item_details = f'DATASET {value.dtype} = {value}'
                        else:
                            item_details = f'{h5_item.shape} {h5_item.dtype}'
                elif isinstance(h5_item, h5py.SoftLink):
                    item_type = 'Soft Link'
                    item_details = f'{h5_item.path}'
                elif isinstance(h5_item, h5py.ExternalLink):
                    item_type = 'External Link'
                    item_details = f'{h5_item.filename}/{h5_item.path}'
                else:
                    logger.debug(f'Unknown item "{item_name}"')

                tree_node.item_data = [item_name, item_type, item_details]

        return root_node


class H5DiffractionFileReader(DiffractionFileReader):
    def __init__(self, data_path: str) -> None:
        self._data_path = data_path
        self._tree_builder = H5DiffractionFileTreeBuilder()

    def read(self, file_path: Path) -> DiffractionDataset:
        with h5py.File(file_path, 'r') as h5_file:
            contents_tree = self._tree_builder.build(h5_file)
            data = h5_file[self._data_path]

            if isinstance(data, h5py.Dataset):
                num_patterns, detector_height, detector_width = data.shape

                metadata = DiffractionMetadata(
                    num_patterns_per_array=[num_patterns],
                    pattern_dtype=data.dtype,
                    detector_extent=ImageExtent(detector_width, detector_height),
                    file_path=file_path,
                )

                array = H5DiffractionPatternArray(
                    label=file_path.stem,
                    indexes=numpy.arange(num_patterns),
                    file_path=file_path,
                    data_path=self._data_path,
                )

                return SimpleDiffractionDataset(metadata, contents_tree, [array])
            else:
                raise ValueError(f'Expected {self._data_path} to be a dataset; got {type(data)}.')


class H5DiffractionFileWriter(DiffractionFileWriter):
    def __init__(self, data_path: str) -> None:
        self._data_path = data_path

    def write(self, file_path: Path, dataset: DiffractionDataset) -> None:
        data = numpy.concatenate([array.get_patterns() for array in dataset])

        with h5py.File(file_path, 'w') as h5_file:
            h5_file.create_dataset(self._data_path, data=data, compression='lzf')


def register_plugins(registry: PluginRegistry) -> None:
    registry.diffraction_file_readers.register_plugin(
        H5DiffractionFileReader(data_path='/exchange/data'),
        simple_name='APS_CSSI',
        display_name='APS 9-ID CSSI Files (*.h5 *.hdf5)',
    )
    registry.diffraction_file_readers.register_plugin(
        H5DiffractionFileReader(data_path='/entry/data/data'),
        simple_name='CNM_APS_HXN',
        display_name='CNM/APS 26-ID Hard X-ray Nanoprobe Files (*.h5 *.hdf5)',
    )
    registry.diffraction_file_readers.register_plugin(
        H5DiffractionFileReader(data_path='/dp'),
        simple_name='fold_slice',
        display_name='fold_slice Files (*.h5 *.hdf5)',
    )
    registry.diffraction_file_writers.register_plugin(
        H5DiffractionFileWriter(data_path='/dp'),
        simple_name='fold_slice',
        display_name='fold_slice Files (*.h5 *.hdf5)',
    )



from .api import ProcessingAPI
from .context import ProcessingProgressMonitor
from .core import ProcessingAlgorithmParameter, ProcessingCore

__all__ = [
    'ProcessingAPI',
    'ProcessingCore',
    'ProcessingProgressMonitor',
    'ProcessingAlgorithmParameter',
]



from __future__ import annotations

from PyQt5.QtCore import Qt
from PyQt5.QtWidgets import (
    QComboBox,
    QDialog,
    QDialogButtonBox,
    QFormLayout,
    QHBoxLayout,
    QMenu,
    QPushButton,
    QTableView,
    QTreeView,
    QVBoxLayout,
    QWidget,
)


class RepositoryButtonBox(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.load_button = QPushButton('Load')
        self.load_menu = QMenu()
        self.save_button = QPushButton('Save')
        self.save_menu = QMenu()
        self.edit_button = QPushButton('Edit')
        self.analyze_button = QPushButton('Analyze')
        self.analyze_menu = QMenu()

        self.load_button.setMenu(self.load_menu)
        self.save_button.setMenu(self.save_menu)
        self.analyze_button.setMenu(self.analyze_menu)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.load_button)
        layout.addWidget(self.save_button)
        layout.addWidget(self.edit_button)
        layout.addWidget(self.analyze_button)
        self.setLayout(layout)


class RepositoryItemCopierDialog(QDialog):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.source_combo_box = QComboBox()
        self.destination_combo_box = QComboBox()
        self.button_box = QDialogButtonBox()

        self.button_box.addButton(QDialogButtonBox.StandardButton.Ok)
        self.button_box.accepted.connect(self.accept)
        self.button_box.addButton(QDialogButtonBox.StandardButton.Cancel)
        self.button_box.rejected.connect(self.reject)

        layout = QFormLayout()
        layout.addRow('From:', self.source_combo_box)
        layout.addRow('To:', self.destination_combo_box)
        layout.addRow(self.button_box)
        self.setLayout(layout)


class RepositoryTableView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.table_view = QTableView()
        self.button_box = RepositoryButtonBox()
        self.copier_dialog = RepositoryItemCopierDialog()

        layout = QVBoxLayout()
        layout.addWidget(self.table_view)
        layout.addWidget(self.button_box)
        self.setLayout(layout)


class RepositoryTreeView(QWidget):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.tree_view = QTreeView()
        self.button_box = RepositoryButtonBox()
        self.copier_dialog = RepositoryItemCopierDialog()

        header = self.tree_view.header()

        if header is not None:
            header.setDefaultAlignment(Qt.AlignmentFlag.AlignCenter)

        layout = QVBoxLayout()
        layout.addWidget(self.tree_view)
        layout.addWidget(self.button_box)
        self.setLayout(layout)



from pathlib import Path
import logging

import lightning
import ptychonn

from .settings import PtychoNNModelSettings, PtychoNNTrainingSettings

logger = logging.getLogger(__name__)


class PtychoNNModelProvider:
    def __init__(
        self,
        model_settings: PtychoNNModelSettings,
        training_settings: PtychoNNTrainingSettings,
        *,
        enable_amplitude: bool,
    ) -> None:
        self._model_settings = model_settings
        self._training_settings = training_settings
        self._enable_amplitude = enable_amplitude
        self._model: ptychonn.LitReconSmallModel | None = None
        self._trainer: lightning.Trainer | None = None

    def get_model_name(self) -> str:
        return 'AmplitudePhase' if self._enable_amplitude else 'PhaseOnly'

    def get_num_channels(self) -> int:
        return 2 if self._enable_amplitude else 1

    def get_model(self) -> ptychonn.LitReconSmallModel:
        if (
            self._model is None
            and self._trainer is not None
            and self._trainer.lightning_module is not None
        ):
            return self._trainer.lightning_module
        else:
            logger.debug('Initializing model from settings')
            self._model = ptychonn.LitReconSmallModel(
                nconv=self._model_settings.num_convolution_kernels.get_value(),
                use_batch_norm=self._model_settings.use_batch_normalization.get_value(),
                enable_amplitude=self._enable_amplitude,
                max_lr=float(self._training_settings.max_learning_rate.get_value()),
                min_lr=float(self._training_settings.min_learning_rate.get_value()),
            )

        return self._model

    def load_model_from_file(self, file_path: Path) -> None:
        logger.debug(f'Reading model from "{file_path}"')
        self._model = ptychonn.LitReconSmallModel.load_from_checkpoint(file_path)
        self._trainer = None

    def set_trainer(self, trainer: lightning.Trainer) -> None:
        self._model = None
        self._trainer = trainer

    def save_model(self, file_path: Path) -> None:
        if self._trainer is None:
            logger.warning('Need trainer to save model!')
        else:
            logger.debug(f'Writing model to "{file_path}"')
            self._trainer.save_checkpoint(file_path)



from collections.abc import Callable, Iterable, Iterator, Mapping
from pathlib import Path
import logging

from ptychodus.api.plugins import PluginChooser
from ptychodus.api.probe import (
    FresnelZonePlate,
    ProbeSequence,
    ProbeFileReader,
    ProbeFileWriter,
)

from ...diffraction import DiffractionAPI
from .average_pattern import AveragePatternProbeBuilder
from .builder import FromFileProbeBuilder, ProbeSequenceBuilder
from .disk import DiskProbeBuilder
from .fzp import FresnelZonePlateProbeBuilder
from .rect import RectangularProbeBuilder
from .settings import ProbeSettings
from .super_gaussian import SuperGaussianProbeBuilder
from .zernike import ZernikeProbeBuilder

logger = logging.getLogger(__name__)


class ProbeBuilderFactory(Iterable[str]):
    def __init__(
        self,
        settings: ProbeSettings,
        diffraction_api: DiffractionAPI,
        fresnel_zone_plate_chooser: PluginChooser[FresnelZonePlate],
        file_reader_chooser: PluginChooser[ProbeFileReader],
        file_writer_chooser: PluginChooser[ProbeFileWriter],
    ) -> None:
        super().__init__()
        self._settings = settings
        self._diffraction_api = diffraction_api
        self._fresnel_zone_plate_chooser = fresnel_zone_plate_chooser
        self._file_reader_chooser = file_reader_chooser
        self._file_writer_chooser = file_writer_chooser
        self._builders: Mapping[str, Callable[[], ProbeSequenceBuilder]] = {
            'disk': lambda: DiskProbeBuilder(settings),
            'average_pattern': self._create_average_pattern_builder,
            'fresnel_zone_plate': self._create_fresnel_zone_plate_builder,
            'rectangular': lambda: RectangularProbeBuilder(settings),
            'super_gaussian': lambda: SuperGaussianProbeBuilder(settings),
            'zernike': lambda: ZernikeProbeBuilder(settings),
        }

    def __iter__(self) -> Iterator[str]:
        return iter(self._builders)

    def create(self, name: str) -> ProbeSequenceBuilder:
        try:
            factory = self._builders[name]
        except KeyError as exc:
            raise KeyError(f'Unknown probe builder "{name}"!') from exc

        return factory()

    def create_default(self) -> ProbeSequenceBuilder:
        return next(iter(self._builders.values()))()

    def create_from_settings(self) -> ProbeSequenceBuilder:
        name = self._settings.builder.get_value()
        name_repaired = name.casefold()

        if name_repaired == 'from_file':
            return self.create_probe_from_file(
                self._settings.file_path.get_value(),
                self._settings.file_type.get_value(),
            )

        return self.create(name_repaired)

    def _create_average_pattern_builder(self) -> ProbeSequenceBuilder:
        return AveragePatternProbeBuilder(self._settings, self._diffraction_api)

    def _create_fresnel_zone_plate_builder(self) -> ProbeSequenceBuilder:
        return FresnelZonePlateProbeBuilder(self._settings, self._fresnel_zone_plate_chooser)

    def get_open_file_filters(self) -> Iterator[str]:
        for plugin in self._file_reader_chooser:
            yield plugin.display_name

    def get_open_file_filter(self) -> str:
        return self._file_reader_chooser.get_current_plugin().display_name

    def create_probe_from_file(self, file_path: Path, file_filter: str) -> ProbeSequenceBuilder:
        self._file_reader_chooser.set_current_plugin(file_filter)
        file_type = self._file_reader_chooser.get_current_plugin().simple_name
        file_reader = self._file_reader_chooser.get_current_plugin().strategy
        return FromFileProbeBuilder(self._settings, file_path, file_type, file_reader)

    def get_save_file_filters(self) -> Iterator[str]:
        for plugin in self._file_writer_chooser:
            yield plugin.display_name

    def get_save_file_filter(self) -> str:
        return self._file_writer_chooser.get_current_plugin().display_name

    def save_probe(self, file_path: Path, file_filter: str, probe: ProbeSequence) -> None:
        self._file_writer_chooser.set_current_plugin(file_filter)
        file_type = self._file_writer_chooser.get_current_plugin().simple_name
        logger.debug(f'Writing "{file_path}" as "{file_type}"')
        file_writer = self._file_writer_chooser.get_current_plugin().strategy
        file_writer.write(file_path, probe)



import logging

import numpy.random

from ptychodus.api.probe_positions import ProbePositionSequence

from .builder import FromMemoryProbePositionsBuilder
from .builder_factory import ProbePositionsBuilderFactory
from .item import ProbePositionsRepositoryItem
from .settings import ProbePositionsSettings
from .transform import ProbePositionTransform

logger = logging.getLogger(__name__)


class ProbePositionsRepositoryItemFactory:
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings: ProbePositionsSettings,
        builder_factory: ProbePositionsBuilderFactory,
    ) -> None:
        self._rng = rng
        self._settings = settings
        self._builder_factory = builder_factory

    def create(self, scan: ProbePositionSequence | None = None) -> ProbePositionsRepositoryItem:
        transform = ProbePositionTransform(self._rng, self._settings)

        if scan is None:
            builder = self._builder_factory.create_default()
        else:
            builder = FromMemoryProbePositionsBuilder(self._settings, scan)
            transform.set_identity()

        return ProbePositionsRepositoryItem(self._settings, builder, transform)

    def create_from_settings(self) -> ProbePositionsRepositoryItem:
        try:
            builder = self._builder_factory.create_from_settings()
        except Exception as exc:
            logger.exception(''.join(exc.args))
            builder = self._builder_factory.create_default()

        transform = ProbePositionTransform(self._rng, self._settings)
        return ProbePositionsRepositoryItem(self._settings, builder, transform)



from pathlib import Path
import logging


from ptychodus.api.geometry import ImageExtent
from ptychodus.api.diffraction import (
    BadPixelsFileReader,
    CropCenter,
    DiffractionFileReader,
    DiffractionFileWriter,
    DiffractionMetadata,
    DiffractionArray,
    SimpleDiffractionDataset,
)
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.reconstructor import AssembledDiffractionData
from ptychodus.api.tree import SimpleTreeNode

from .bad_pixels import BadPixelsProvider
from .dataset import AssembledDiffractionDataset
from .settings import DetectorSettings, DiffractionSettings

logger = logging.getLogger(__name__)


class PatternsStreamingContext:
    def __init__(self, dataset: AssembledDiffractionDataset, metadata: DiffractionMetadata) -> None:
        self._dataset = dataset
        self._metadata = metadata

    def start(self) -> None:
        contents_tree = SimpleTreeNode.create_root(['Name', 'Type', 'Details'])
        stream_dataset = SimpleDiffractionDataset(self._metadata, contents_tree, [])
        self._dataset.reload(stream_dataset)
        self._dataset.load_all_arrays(process_patterns=True, block=True)

    def append_array(self, array: DiffractionArray) -> None:
        self._dataset.append_array(array, process_patterns=True)

    def stop(self) -> None:
        pass


class DiffractionAPI:
    def __init__(
        self,
        diffraction_settings: DiffractionSettings,
        detector_settings: DetectorSettings,
        bad_pixels_provider: BadPixelsProvider,
        dataset: AssembledDiffractionDataset,
        bad_pixels_file_reader_chooser: PluginChooser[BadPixelsFileReader],
        file_reader_chooser: PluginChooser[DiffractionFileReader],
        file_writer_chooser: PluginChooser[DiffractionFileWriter],
    ) -> None:
        super().__init__()
        self._diffraction_settings = diffraction_settings
        self._detector_settings = detector_settings
        self._bad_pixels_provider = bad_pixels_provider
        self._dataset = dataset
        self._bad_pixels_file_reader_chooser = bad_pixels_file_reader_chooser
        self._file_reader_chooser = file_reader_chooser
        self._file_writer_chooser = file_writer_chooser

    def create_streaming_context(self, metadata: DiffractionMetadata) -> PatternsStreamingContext:
        return PatternsStreamingContext(self._dataset, metadata)

    def get_bad_pixels_file_reader_chooser(self) -> PluginChooser[BadPixelsFileReader]:
        return self._bad_pixels_file_reader_chooser

    def open_bad_pixels(self, file_path: Path, *, file_type: str | None = None) -> None:
        if file_path.is_file():
            if file_type is not None:
                self._bad_pixels_file_reader_chooser.set_current_plugin(file_type)

            plugin = self._bad_pixels_file_reader_chooser.get_current_plugin()
            logger.debug(f'Reading "{file_path}" as "{plugin.simple_name}"')

            try:
                bad_pixels = plugin.strategy.read(file_path)
            except Exception as exc:
                raise RuntimeError(f'Failed to read "{file_path}"') from exc
            else:
                self._bad_pixels_provider.set_bad_pixels(bad_pixels)
                self._detector_settings.bad_pixels_file_path.set_value(file_path)
        else:
            logger.warning(f'Refusing to read invalid file path {file_path}')

    def clear_bad_pixels(self) -> None:
        self._bad_pixels_provider.clear_bad_pixels()
        self._detector_settings.bad_pixels_file_path.set_value(Path('/path/to/bad_pixels.npy'))

    def get_file_reader_chooser(self) -> PluginChooser[DiffractionFileReader]:
        return self._file_reader_chooser

    def open_patterns(
        self,
        file_path: Path,
        *,
        file_type: str | None = None,
        crop_center: CropCenter | None = None,
        crop_extent: ImageExtent | None = None,
        detector_extent: ImageExtent | None = None,
        process_patterns: bool = True,
        block: bool = False,
    ) -> int:
        if crop_center is not None:
            self._diffraction_settings.crop_center_x_px.set_value(crop_center.position_x_px)
            self._diffraction_settings.crop_center_y_px.set_value(crop_center.position_y_px)

        if crop_extent is not None:
            self._diffraction_settings.crop_width_px.set_value(crop_extent.width_px)
            self._diffraction_settings.crop_height_px.set_value(crop_extent.height_px)

        if detector_extent is not None:
            self._detector_settings.width_px.set_value(detector_extent.width_px)
            self._detector_settings.height_px.set_value(detector_extent.height_px)

        if file_path.is_file():
            if file_type is not None:
                self._file_reader_chooser.set_current_plugin(file_type)

            plugin = self._file_reader_chooser.get_current_plugin()
            logger.debug(f'Reading "{file_path}" as "{plugin.simple_name}"')

            try:
                dataset = plugin.strategy.read(file_path)
            except Exception as exc:
                raise RuntimeError(f'Failed to read "{file_path}"') from exc
            else:
                self._dataset.reload(dataset)

                if block:
                    self._dataset.load_all_arrays(process_patterns=process_patterns, block=True)

                return 0
        else:
            logger.warning(f'Refusing to read invalid file path {file_path}')

        return -1

    def load_all_arrays(self, *, process_patterns: bool = True, block: bool = False) -> None:
        self._dataset.load_all_arrays(process_patterns=process_patterns, block=block)

    def close_patterns(self) -> None:
        self._dataset.clear()

    def get_file_writer_chooser(self) -> PluginChooser[DiffractionFileWriter]:
        return self._file_writer_chooser

    def save_patterns(self, file_path: Path, file_type: str) -> None:
        self._file_writer_chooser.set_current_plugin(file_type)
        file_type = self._file_writer_chooser.get_current_plugin().simple_name
        logger.debug(f'Writing "{file_path}" as "{file_type}"')
        writer = self._file_writer_chooser.get_current_plugin().strategy
        writer.write(file_path, self._dataset)

    def get_assembled_data(self) -> AssembledDiffractionData:
        return self._dataset.get_assembled_data()

    def import_assembled_patterns(self, file_path: Path) -> None:
        self._dataset.import_assembled_patterns(file_path)

    def export_assembled_patterns(self, file_path: Path) -> None:
        self._dataset.export_assembled_patterns(file_path)



import numpy

from ptychodus.api.object import ObjectFileReader, ObjectFileWriter
from ptychodus.api.observer import Observable, Observer
from ptychodus.api.plugins import PluginChooser
from ptychodus.api.probe import FresnelZonePlate, ProbeFileReader, ProbeFileWriter
from ptychodus.api.product import ProductFileReader, ProductFileWriter
from ptychodus.api.probe_positions import ProbePositionFileReader, ProbePositionFileWriter
from ptychodus.api.settings import SettingsRegistry

from ..diffraction import AssembledDiffractionDataset, DiffractionAPI, PatternSizer
from .api import ObjectAPI, ProbeAPI, ProductAPI, ProbePositionsAPI
from .item_factory import ProductRepositoryItemFactory
from .object import ObjectBuilderFactory, ObjectRepositoryItemFactory, ObjectSettings
from .object_repository import ObjectRepository
from .probe import ProbeBuilderFactory, ProbeRepositoryItemFactory, ProbeSettings
from .probe_repository import ProbeRepository
from .repository import ProductRepository
from .probe_positions import (
    ProbePositionsBuilderFactory,
    ProbePositionsRepositoryItemFactory,
    ProbePositionsSettings,
)
from .scan_repository import ProbePositionsRepository
from .settings import ProductSettings


class ProductCore(Observer):
    def __init__(
        self,
        rng: numpy.random.Generator,
        settings_registry: SettingsRegistry,
        pattern_sizer: PatternSizer,
        diffraction_api: DiffractionAPI,
        diffraction_dataset: AssembledDiffractionDataset,
        scan_file_reader_chooser: PluginChooser[ProbePositionFileReader],
        scan_file_writer_chooser: PluginChooser[ProbePositionFileWriter],
        fresnel_zone_plate_chooser: PluginChooser[FresnelZonePlate],
        probe_file_reader_chooser: PluginChooser[ProbeFileReader],
        probe_file_writer_chooser: PluginChooser[ProbeFileWriter],
        object_file_reader_chooser: PluginChooser[ObjectFileReader],
        object_file_writer_chooser: PluginChooser[ObjectFileWriter],
        product_file_reader_chooser: PluginChooser[ProductFileReader],
        product_file_writer_chooser: PluginChooser[ProductFileWriter],
        reinit_observable: Observable,
    ) -> None:
        super().__init__()
        self.settings = ProductSettings(settings_registry)

        self._scan_settings = ProbePositionsSettings(settings_registry)
        self._scan_builder_factory = ProbePositionsBuilderFactory(
            self._scan_settings, scan_file_reader_chooser, scan_file_writer_chooser
        )
        self._scan_repository_item_factory = ProbePositionsRepositoryItemFactory(
            rng, self._scan_settings, self._scan_builder_factory
        )

        self._probe_settings = ProbeSettings(settings_registry)
        self._probe_builder_factory = ProbeBuilderFactory(
            self._probe_settings,
            diffraction_api,
            fresnel_zone_plate_chooser,
            probe_file_reader_chooser,
            probe_file_writer_chooser,
        )
        self._probe_repository_item_factory = ProbeRepositoryItemFactory(
            rng, self._probe_settings, self._probe_builder_factory
        )

        self._object_settings = ObjectSettings(settings_registry)
        self._object_builder_factory = ObjectBuilderFactory(
            rng,
            self._object_settings,
            diffraction_api,
            object_file_reader_chooser,
            object_file_writer_chooser,
        )
        self._object_repository_item_factory = ObjectRepositoryItemFactory(
            rng, self._object_settings, self._object_builder_factory
        )

        self.product_repository = ProductRepository()
        self._item_factory = ProductRepositoryItemFactory(
            self.settings,
            pattern_sizer,
            diffraction_dataset,
            self._scan_repository_item_factory,
            self._probe_repository_item_factory,
            self._object_repository_item_factory,
            self.product_repository,
            product_file_reader_chooser,
        )
        self.product_api = ProductAPI(
            self.settings,
            self.product_repository,
            self._item_factory,
            product_file_reader_chooser,
            product_file_writer_chooser,
        )
        self.probe_positions_repository = ProbePositionsRepository(self.product_repository)
        self.probe_positions_api = ProbePositionsAPI(
            self._scan_settings, self.probe_positions_repository, self._scan_builder_factory
        )
        self.probe_repository = ProbeRepository(self.product_repository)
        self.probe_api = ProbeAPI(
            self._probe_settings, self.probe_repository, self._probe_builder_factory
        )
        self.object_repository = ObjectRepository(self.product_repository)
        self.object_api = ObjectAPI(
            self._object_settings, self.object_repository, self._object_builder_factory
        )

        # TODO vvv refactor vvv
        product_file_reader_chooser.synchronize_with_parameter(self.settings.file_type)
        product_file_writer_chooser.set_current_plugin(self.settings.file_type.get_value())
        scan_file_reader_chooser.synchronize_with_parameter(self._scan_settings.file_type)
        scan_file_writer_chooser.set_current_plugin(self._scan_settings.file_type.get_value())
        probe_file_reader_chooser.synchronize_with_parameter(self._probe_settings.file_type)
        probe_file_writer_chooser.set_current_plugin(self._probe_settings.file_type.get_value())
        object_file_reader_chooser.synchronize_with_parameter(self._object_settings.file_type)
        object_file_writer_chooser.set_current_plugin(self._object_settings.file_type.get_value())
        # TODO ^^^^^^^^^^^^^^^^

        self._reinit_observable = reinit_observable
        reinit_observable.add_observer(self)

    def _update(self, observable: Observable) -> None:
        if observable is self._reinit_observable:
            self.product_api.insert_product_from_settings()



import logging

import numpy

from ptychodus.api.diffraction import BadPixels
from ptychodus.api.geometry import ImageExtent
from ptychodus.api.observer import Observable, Observer

from .settings import DetectorSettings

logger = logging.getLogger(__name__)


class BadPixelsProvider(Observable, Observer):
    def __init__(self, settings: DetectorSettings) -> None:
        super().__init__()
        self._settings = settings
        self._bad_pixels: BadPixels | None = None

        settings.height_px.add_observer(self)
        settings.width_px.add_observer(self)

    def set_detector_extent(self, extent: ImageExtent) -> None:
        self._settings.height_px.set_value(extent.height_px)
        self._settings.width_px.set_value(extent.width_px)

    def clear_bad_pixels(self) -> None:
        self._bad_pixels = None
        self.notify_observers()

    def set_bad_pixels(self, bad_pixels: BadPixels) -> None:
        if bad_pixels.ndim != 2:
            raise ValueError(f'Bad pixels array must be 2D, got {bad_pixels.ndim}D.')

        num_bad_pixels = int(numpy.count_nonzero(bad_pixels))
        self._bad_pixels = bad_pixels if num_bad_pixels > 0 else None
        self._settings.height_px.set_value(bad_pixels.shape[-2])
        self._settings.width_px.set_value(bad_pixels.shape[-1])
        self.notify_observers()

    def get_bad_pixels(self) -> BadPixels:
        if self._bad_pixels is None:
            height_px = self._settings.height_px.get_value()
            width_px = self._settings.width_px.get_value()
            return numpy.full((height_px, width_px), False)

        return self._bad_pixels

    def get_num_bad_pixels(self) -> int:
        if self._bad_pixels is None:
            return 0

        return int(numpy.count_nonzero(self._bad_pixels))

    def _clear_bad_pixels_if_mismatched(self, actual: int, expected: int) -> None:
        if actual != expected:
            self.clear_bad_pixels()

    def _update(self, observable: Observable) -> None:
        if self._bad_pixels is not None:
            match observable:
                case self._settings.height_px:
                    self._clear_bad_pixels_if_mismatched(
                        self._bad_pixels.shape[-2], self._settings.height_px.get_value()
                    )
                case self._settings.width_px:
                    self._clear_bad_pixels_if_mismatched(
                        self._bad_pixels.shape[-1], self._settings.width_px.get_value()
                    )



from pathlib import Path
import logging

import h5py
import numpy

from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe_positions import (
    ProbePositionSequence,
    ProbePositionFileReader,
    ProbePosition,
    ProbePositionParseError,
)

logger = logging.getLogger(__name__)


class FoldSlicePositionFileReader(ProbePositionFileReader):
    def read(self, file_path: Path) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        with h5py.File(file_path, 'r') as h5_file:
            try:
                pp_x = numpy.squeeze(h5_file['/ppX'])
                pp_y = numpy.squeeze(h5_file['/ppY'])
            except KeyError:
                logger.warning('Unable to find data.')
            else:
                if pp_x.shape == pp_y.shape:
                    logger.debug(f'Coordinate arrays have shape {pp_x.shape}.')
                else:
                    raise ProbePositionParseError('Coordinate array shape mismatch!')

                for idx, (x, y) in enumerate(zip(pp_x, pp_y)):
                    point = ProbePosition(idx, x, y)
                    point_list.append(point)

        return ProbePositionSequence(point_list)


def register_plugins(registry: PluginRegistry) -> None:
    registry.probe_position_file_readers.register_plugin(
        FoldSlicePositionFileReader(),
        simple_name='fold_slice',
        display_name='fold_slice Files (*.h5 *.hdf5)',
    )



from collections.abc import Sequence
from typing import overload
import logging

logger = logging.getLogger(__name__)


class PtyChiDeviceRepository(Sequence[str]):
    def __init__(self, *, is_developer_mode_enabled: bool) -> None:
        self._devices: list[str] = list()

        try:
            import ptychi
        except ModuleNotFoundError:
            if is_developer_mode_enabled:
                self._devices.extend(f'gpu:{n}' for n in range(4))
        else:
            for device in ptychi.list_available_devices():
                logger.info(device)
                self._devices.append(f'{device.name} ({device.torch_device})')

        if not self._devices:
            logger.info('No devices found!')

    @overload
    def __getitem__(self, index: int) -> str: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[str]: ...

    def __getitem__(self, index: int | slice) -> str | Sequence[str]:
        return self._devices[index]

    def __len__(self) -> int:
        return len(self._devices)



from __future__ import annotations
import logging

from PyQt5.QtCore import PYQT_VERSION_STR, QSize, QT_VERSION_STR, Qt
from PyQt5.QtGui import QIcon
from PyQt5.QtWidgets import (
    QActionGroup,
    QApplication,
    QLCDNumber,
    QMainWindow,
    QSizePolicy,
    QSplitter,
    QStackedWidget,
    QTableView,
    QToolBar,
    QWidget,
)

from . import resources  # noqa
from .agent import AgentView, AgentChatView
from .diffraction import PatternsView
from .image import ImageView
from .product import ProductView
from .processing import ProcessingStatusView
from .repository import RepositoryTableView, RepositoryTreeView
from .probe_positions import ProbePositionsPlotView
from .settings import SettingsView

logger = logging.getLogger(__name__)


class ViewCore(QMainWindow):
    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)

        logger.info(f'PyQt {PYQT_VERSION_STR}')
        logger.info(f'Qt {QT_VERSION_STR}')

        self.navigation_tool_bar = QToolBar()
        self.navigation_action_group = QActionGroup(self.navigation_tool_bar)

        self.splitter = QSplitter(Qt.Orientation.Horizontal)
        self.left_panel = QStackedWidget()
        self.right_panel = QStackedWidget()
        self.memory_widget = QLCDNumber()

        self.settings_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/settings'), 'Settings'
        )
        self.settings_view = SettingsView()
        self.settings_table_view = QTableView()

        self.patterns_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/patterns'), 'Patterns'
        )
        self.patterns_view = PatternsView()
        self.patterns_image_view = ImageView()

        self.product_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/products'), 'Products'
        )
        self.product_view = ProductView()
        self.product_diagram_view = QWidget()

        self.positions_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/positions'), 'Positions'
        )
        self.probe_positions_view = RepositoryTableView()
        self.probe_positions_plot_view = ProbePositionsPlotView()

        self.probe_action = self.navigation_tool_bar.addAction(QIcon(':/icons/probe'), 'Probe')
        self.probe_view = RepositoryTreeView()
        self.probe_image_view = ImageView()

        self.object_action = self.navigation_tool_bar.addAction(QIcon(':/icons/object'), 'Object')
        self.object_view = RepositoryTreeView()
        self.object_image_view = ImageView()

        self.processing_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/processing'), 'Processing'
        )
        self.processing_view = QWidget()
        self.processing_status_view = ProcessingStatusView()

        self.globus_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/globus'),
            'Globus',
        )
        self.globus_view = QWidget()
        self.globus_status_view = QTableView()

        self.automation_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/automate'), 'Automation'
        )
        self.automation_view = QWidget()
        self.automation_widget = QWidget()

        self.agent_action = self.navigation_tool_bar.addAction(
            QIcon(':/icons/sparkles'),
            'Agent',
        )
        self.agent_view = AgentView()
        self.agent_chat_view = AgentChatView()

        #####

        self.setWindowIcon(QIcon(':/icons/ptychodus'))

        self.navigation_tool_bar.setContextMenuPolicy(Qt.ContextMenuPolicy.PreventContextMenu)
        self.navigation_tool_bar.setToolButtonStyle(Qt.ToolButtonStyle.ToolButtonTextUnderIcon)
        self.navigation_tool_bar.setIconSize(QSize(32, 32))
        self.addToolBar(Qt.ToolBarArea.LeftToolBarArea, self.navigation_tool_bar)

        for index, action in enumerate(self.navigation_tool_bar.actions()):
            action.setCheckable(True)
            action.setData(index)
            self.navigation_action_group.addAction(action)

        # maintain same order as navigationToolBar buttons
        self.left_panel.addWidget(self.settings_view)
        self.left_panel.addWidget(self.patterns_view)
        self.left_panel.addWidget(self.product_view)
        self.left_panel.addWidget(self.probe_positions_view)
        self.left_panel.addWidget(self.probe_view)
        self.left_panel.addWidget(self.object_view)
        self.left_panel.addWidget(self.processing_view)
        self.left_panel.addWidget(self.globus_view)
        self.left_panel.addWidget(self.automation_view)
        self.left_panel.addWidget(self.agent_view)
        self.left_panel.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self.splitter.addWidget(self.left_panel)

        # maintain same order as navigationToolBar buttons
        self.right_panel.addWidget(self.settings_table_view)
        self.right_panel.addWidget(self.patterns_image_view)
        self.right_panel.addWidget(self.product_diagram_view)
        self.right_panel.addWidget(self.probe_positions_plot_view)
        self.right_panel.addWidget(self.probe_image_view)
        self.right_panel.addWidget(self.object_image_view)
        self.right_panel.addWidget(self.processing_status_view)
        self.right_panel.addWidget(self.globus_status_view)
        self.right_panel.addWidget(self.automation_widget)
        self.right_panel.addWidget(self.agent_chat_view)
        self.right_panel.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Minimum)
        self.splitter.addWidget(self.right_panel)

        self.setCentralWidget(self.splitter)

        application_desktop = QApplication.desktop()

        if application_desktop is not None:
            desktop_size = application_desktop.availableGeometry().size()
            preferred_height = desktop_size.height() * 2 // 3
            preferred_width = min(desktop_size.width() * 2 // 3, 2 * preferred_height)
            self.resize(preferred_width, preferred_height)

        status_bar = self.statusBar()

        if status_bar is not None:
            status_bar.addPermanentWidget(self.memory_widget)



from __future__ import annotations
from typing import Any, overload

from PyQt5.QtCore import Qt, QAbstractItemModel, QModelIndex, QObject

from ptychodus.api.common import BYTES_PER_MEGABYTE
from ptychodus.api.diffraction import DiffractionPattern

from ptychodus.model.diffraction import AssembledDiffractionArray

__all__ = ['DatasetTreeModel']


class DatasetTreeNode:
    def __init__(
        self,
        parent_node: DatasetTreeNode | None,
        array: AssembledDiffractionArray,
        frame_index: int,
    ) -> None:
        self.parent_node = parent_node
        self._array = array
        self._frame_index = frame_index
        self.child_nodes: list[DatasetTreeNode] = list()

    @classmethod
    def create_root(cls) -> DatasetTreeNode:
        return cls(None, AssembledDiffractionArray.create_null(), -1)

    def insert_child(self, pos: int, array: AssembledDiffractionArray) -> DatasetTreeNode:
        child = DatasetTreeNode(self, array, -1)

        for frame_index in range(array.get_num_patterns()):
            grandchild = DatasetTreeNode(child, array, frame_index)
            child.child_nodes.append(grandchild)

        self.child_nodes.insert(pos, child)
        return child

    def get_label(self) -> str:
        return self._array.get_label() if self._frame_index < 0 else f'Frame {self._frame_index}'

    def get_data(self) -> DiffractionPattern:
        return (
            self._array.get_average_pattern()
            if self._frame_index < 0
            else self._array.get_pattern(self._frame_index)
        )

    def get_counts(self) -> int:
        return (
            int(self._array.get_mean_pattern_counts())
            if self._frame_index < 0
            else int(self._array.get_pattern_counts(self._frame_index))
        )

    def get_nframes(self) -> int:
        return len(self.child_nodes) if self._frame_index < 0 else 1

    def get_nbytes(self) -> int:
        return (
            self._array.get_patterns().nbytes
            if self._frame_index < 0
            else self._array.get_pattern(self._frame_index).nbytes
        )

    def get_row(self) -> int:
        return 0 if self.parent_node is None else self.parent_node.child_nodes.index(self)


class DatasetTreeModel(QAbstractItemModel):
    def __init__(self, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._nodes = DatasetTreeNode.create_root()
        self._max_counts = 1
        self._header = ['Label', 'Counts', 'Frames', 'Size [MB]']

    def clear(self) -> None:
        self.beginResetModel()
        self._nodes = DatasetTreeNode.create_root()
        self._max_counts = 1
        self.endResetModel()

    def insert_array(self, row: int, array: AssembledDiffractionArray) -> None:
        max_counts = array.get_max_pattern_counts()

        if self._max_counts < max_counts:
            self._max_counts = max_counts
            num_rows = self.rowCount()

            top_left = self.index(0, 1)
            bottom_right = self.index(num_rows - 1, 1)
            self.dataChanged.emit(top_left, bottom_right)

            for row2 in range(num_rows):
                parent_index = self.index(row2, 0)
                num_rows2 = self.rowCount(parent_index)

                child_top_left = self.index(0, 1, parent_index)
                child_bottom_right = self.index(num_rows2 - 1, 1, parent_index)
                self.dataChanged.emit(child_top_left, child_bottom_right)

        self.beginInsertRows(QModelIndex(), row, row)
        child_node = self._nodes.insert_child(row, array)
        self.endInsertRows()

        index = self.index(row, 0)
        self.beginInsertRows(index, 0, len(child_node.child_nodes))
        self.endInsertRows()

    def refresh_array(self, row: int) -> None:
        top_left = self.index(row, 0)
        bottom_right = self.index(row, self.columnCount() - 1)
        self.dataChanged.emit(top_left, bottom_right)

        num_rows = self.rowCount(top_left)
        num_cols = self.columnCount(top_left)

        child_top_left = self.index(0, 0, top_left)
        child_bottom_right = self.index(num_rows - 1, num_cols - 1, top_left)
        self.dataChanged.emit(child_top_left, child_bottom_right)

    @overload
    def parent(self, child: QModelIndex) -> QModelIndex: ...

    @overload
    def parent(self) -> QObject: ...

    def parent(self, child: QModelIndex | None = None) -> QModelIndex | QObject:
        if child is None:
            return super().parent()

        if child.isValid():
            child_node = child.internalPointer()
            parent_node = child_node.parent_node

            if parent_node is not self._nodes:
                return self.createIndex(parent_node.get_row(), 0, parent_node)

        return QModelIndex()

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[section]

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid():
            node = index.internalPointer()

            if role == Qt.ItemDataRole.DisplayRole:
                match index.column():
                    case 0:
                        return node.get_label()
                    case 1:
                        return str(node.get_counts())
                    case 2:
                        return node.get_nframes()
                    case 3:
                        return f'{node.get_nbytes() / BYTES_PER_MEGABYTE:.2f}'
            elif role == Qt.ItemDataRole.UserRole:
                if index.column() == 1:
                    return int(100 * node.get_counts()) // int(self._max_counts)

    def index(self, row: int, column: int, parent: QModelIndex = QModelIndex()) -> QModelIndex:
        if self.hasIndex(row, column, parent):
            parent_node = parent.internalPointer() if parent.isValid() else self._nodes
            child_node = parent_node.child_nodes[row]

            if child_node:
                return self.createIndex(row, column, child_node)

        return QModelIndex()

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        node = parent.internalPointer() if parent.isValid() else self._nodes
        return len(node.child_nodes)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._header)



"""Diffraction data structures and file I/O plugin interfaces."""

from __future__ import annotations
from abc import ABC, abstractmethod
from collections.abc import Sequence
from dataclasses import dataclass
from pathlib import Path
from typing import overload, Any, TypeAlias

import numpy

from .geometry import ImageExtent, PixelGeometry
from .tree import SimpleTreeNode

BadPixels: TypeAlias = numpy.ndarray[tuple[int, int], numpy.dtype[numpy.bool_]]
DiffractionPatternDType: TypeAlias = numpy.dtype[numpy.integer[Any] | numpy.floating[Any]]
DiffractionPatternCounts: TypeAlias = numpy.ndarray[tuple[int], DiffractionPatternDType]
DiffractionPattern: TypeAlias = numpy.ndarray[tuple[int, int], DiffractionPatternDType]
DiffractionPatterns: TypeAlias = numpy.ndarray[tuple[int, int, int], DiffractionPatternDType]
DiffractionIndexes: TypeAlias = numpy.ndarray[tuple[int], numpy.dtype[numpy.integer[Any]]]


@dataclass(frozen=True)
class CropCenter:
    """Pixel coordinates of the center used when cropping diffraction patterns."""

    position_x_px: int
    position_y_px: int


class DiffractionArray(ABC):
    """A block of diffraction patterns with associated scan indexes."""

    @abstractmethod
    def get_label(self) -> str:
        pass

    @abstractmethod
    def get_indexes(self) -> DiffractionIndexes:
        pass

    @abstractmethod
    def get_patterns(self) -> DiffractionPatterns:
        pass

    def get_num_patterns(self) -> int:
        return self.get_patterns().shape[0]


class SimpleDiffractionArray(DiffractionArray):
    """Concrete DiffractionArray backed by in-memory numpy arrays."""

    def __init__(
        self,
        label: str,
        indexes: DiffractionIndexes,
        patterns: DiffractionPatterns,
    ) -> None:
        super().__init__()
        self._label = label
        self._indexes = indexes
        self._patterns = patterns

    def get_label(self) -> str:
        return self._label

    def get_indexes(self) -> DiffractionIndexes:
        return self._indexes

    def get_patterns(self) -> DiffractionPatterns:
        return self._patterns


@dataclass(frozen=True)
class DiffractionMetadata:
    """Metadata describing a diffraction dataset (geometry, energy, file path, etc.)."""

    num_patterns_per_array: Sequence[int]
    pattern_dtype: DiffractionPatternDType
    detector_distance_m: float | None = None
    detector_extent: ImageExtent | None = None
    detector_pixel_geometry: PixelGeometry | None = None
    detector_bit_depth: int | None = None
    crop_center: CropCenter | None = None
    probe_energy_eV: float | None = None  # noqa: N815
    probe_photon_count: int | None = None
    exposure_time_s: float | None = None
    tomography_angle_deg: float | None = None
    file_path: Path | None = None

    @classmethod
    def create_null(cls, file_path: Path | None = None) -> DiffractionMetadata:
        return cls([], numpy.dtype(numpy.ubyte), file_path=file_path)


class DiffractionDataset(Sequence[DiffractionArray], ABC):
    """A sequence of DiffractionArrays with shared metadata and bad-pixel mask."""

    @abstractmethod
    def get_metadata(self) -> DiffractionMetadata:
        pass

    @abstractmethod
    def get_layout(self) -> SimpleTreeNode:
        pass

    @abstractmethod
    def get_bad_pixels(self) -> BadPixels | None:
        pass


class SimpleDiffractionDataset(DiffractionDataset):
    """Concrete DiffractionDataset backed by a list of DiffractionArray objects."""

    def __init__(
        self,
        metadata: DiffractionMetadata,
        contents_tree: SimpleTreeNode,
        array_list: Sequence[DiffractionArray],
        bad_pixels: BadPixels | None = None,
    ) -> None:
        super().__init__()
        self._metadata = metadata
        self._contents_tree = contents_tree
        self._array_list = array_list
        self._bad_pixels = bad_pixels

    @classmethod
    def create_null(cls, file_path: Path | None = None) -> SimpleDiffractionDataset:
        metadata = DiffractionMetadata.create_null(file_path)
        contents_tree = SimpleTreeNode.create_root(list())
        array_list: list[DiffractionArray] = list()
        return cls(metadata, contents_tree, array_list)

    def get_metadata(self) -> DiffractionMetadata:
        return self._metadata

    def get_layout(self) -> SimpleTreeNode:
        return self._contents_tree

    def get_bad_pixels(self) -> BadPixels | None:
        return self._bad_pixels

    @overload
    def __getitem__(self, index: int) -> DiffractionArray: ...

    @overload
    def __getitem__(self, index: slice) -> Sequence[DiffractionArray]: ...

    def __getitem__(self, index: int | slice) -> DiffractionArray | Sequence[DiffractionArray]:
        return self._array_list[index]

    def __len__(self) -> int:
        return len(self._array_list)


class DiffractionFileReader(ABC):
    """Plugin interface for reading diffraction files."""

    @abstractmethod
    def read(self, file_path: Path) -> DiffractionDataset:
        """Read a diffraction dataset from file."""
        pass


class DiffractionFileWriter(ABC):
    """Plugin interface for writing diffraction files."""

    @abstractmethod
    def write(self, file_path: Path, dataset: DiffractionDataset) -> None:
        """Write a diffraction dataset to file."""
        pass


class BadPixelsFileReader(ABC):
    """Plugin interface for reading bad-pixel mask files."""

    @abstractmethod
    def read(self, file_path: Path) -> BadPixels:
        """Read a bad-pixel mask from file."""
        pass



from typing import Final

from PyQt5.QtWidgets import QFrame, QLCDNumber, QSizePolicy

from ..model.memory import MemoryPresenter


class MemoryController:
    UPDATE_INTERVAL_S: Final[int] = 10

    def __init__(self, presenter: MemoryPresenter, widget: QLCDNumber) -> None:
        self._presenter = presenter
        self._widget = widget
        self._widget.setSegmentStyle(QLCDNumber.SegmentStyle.Flat)
        self._widget.setFrameStyle(QFrame.Shape.Panel | QFrame.Shadow.Sunken)
        self._widget.setDigitCount(6)
        self._widget.setSizePolicy(QSizePolicy.Policy.Minimum, QSizePolicy.Policy.Preferred)

    def run_tasks(self, one_second_counter: int) -> None:
        if one_second_counter % MemoryController.UPDATE_INTERVAL_S != 0:
            return

        stats = self._presenter.get_statistics()
        total_MB = int(stats.total_physical_memory_bytes / 1e6)  # noqa: N806
        total_str = f'Total Memory: {total_MB} MB'

        avail_MB = int(stats.available_memory_bytes / 1e6)  # noqa: N806
        avail_str = f'Available Memory: {avail_MB} MB'

        self._widget.display(avail_MB)
        self._widget.setToolTip('\n'.join((total_str, avail_str)))



from ptychodus.api.plugins import PluginRegistry

from .aps33id_velociprobe_diffraction_file import VelociprobeDiffractionFileReader
from .aps33id_velociprobe_position_file import VelociprobePositionFileReader


def register_plugins(registry: PluginRegistry) -> None:
    diffraction_file_reader = VelociprobeDiffractionFileReader()

    registry.diffraction_file_readers.register_plugin(
        diffraction_file_reader,
        simple_name='APS_Velociprobe',
        display_name='APS 33-ID Velociprobe Files (*.h5 *.hdf5)',
    )
    registry.probe_position_file_readers.register_plugin(
        VelociprobePositionFileReader.create_laser_interferometer_instance(diffraction_file_reader),
        simple_name='APS_Velociprobe_LI',
        display_name='APS 33-ID Velociprobe Files - Laser Interferometer (*.txt)',
    )
    registry.probe_position_file_readers.register_plugin(
        VelociprobePositionFileReader.create_position_encoder_instance(diffraction_file_reader),
        simple_name='APS_Velociprobe_PE',
        display_name='APS 33-ID Velociprobe Files - Position Encoder (*.txt)',
    )



from importlib.metadata import version, PackageNotFoundError
from typing import Final

try:
    __version__ = version('ptychodus')
except PackageNotFoundError:
    # package is not installed
    __version__ = 'unknown'

try:
    from .ptychodus_stream_processor import PtychodusAdImageProcessor
except ModuleNotFoundError:
    pass

__all__ = [
    'api',
    'cli',
    'model',
    'view',
    'controller',
    'PtychodusAdImageProcessor',
    'VERSION_STRING',
]

VERSION_STRING: Final[str] = f'Ptychodus ({__version__})'



from pathlib import Path
import logging

from ptychodus.api.io import StandardFileLayout
from ptychodus.api.settings import SettingsRegistry

from ..diffraction import DiffractionAPI
from ..product import ProductAPI
from ..processing import ProcessingAPI
from .client import GlobusClient, GlobusJob
from .settings import GlobusSettings

logger = logging.getLogger(__name__)


class GlobusExecutor:
    def __init__(
        self,
        settings: GlobusSettings,
        settings_registry: SettingsRegistry,
        diffraction_api: DiffractionAPI,
        product_api: ProductAPI,
        processing_api: ProcessingAPI,
        client: GlobusClient,
    ) -> None:
        super().__init__()
        self._settings = settings
        self._settings_registry = settings_registry
        self._diffraction_api = diffraction_api
        self._product_api = product_api
        self._processing_api = processing_api
        self._client = client

    def populate_input_directory(self, input_product_index: int) -> Path:
        try:
            product_item = self._product_api.get_item(input_product_index)
        except IndexError:
            logger.exception(f'Failed access product for flow ({input_product_index=})!')
            raise

        input_directory = self._settings.input_data_posix_path.get_value() / product_item.get_name()

        try:
            input_directory.mkdir(mode=0o755, parents=True, exist_ok=True)
        except FileExistsError:
            logger.exception('Input data POSIX path must be a directory!')
            raise

        self._settings_registry.save_settings(input_directory / StandardFileLayout.SETTINGS)
        self._diffraction_api.export_assembled_patterns(
            input_directory / StandardFileLayout.DIFFRACTION
        )
        self._product_api.save_product(
            input_product_index,
            input_directory / StandardFileLayout.PRODUCT_IN,
            file_type='HDF5',
        )

        return input_directory

    def _run_flow(self, ptychodus_action: str, flow_label: str) -> None:
        input_data_globus_path = f'{self._settings.input_data_globus_path.get_value()}/{flow_label}'

        compute_data_posix_path = self._settings.compute_data_posix_path.get_value() / flow_label
        compute_data_globus_path = (
            f'{self._settings.compute_data_globus_path.get_value()}/{flow_label}'
        )
        output_data_globus_path = (
            f'{self._settings.output_data_globus_path.get_value()}/{flow_label}'
        )

        flow_input = {
            'input_data_transfer_source_endpoint': str(
                self._settings.input_data_endpoint_id.get_value()
            ),
            'input_data_transfer_source_path': input_data_globus_path,
            'input_data_transfer_destination_endpoint': str(
                self._settings.compute_data_endpoint_id.get_value()
            ),
            'input_data_transfer_destination_path': compute_data_globus_path,
            'input_data_transfer_recursive': True,
            'input_data_transfer_sync_level': self._settings.transfer_sync_level.get_value(),
            'compute_endpoint': str(self._settings.compute_endpoint_id.get_value()),
            'ptychodus_action': ptychodus_action,
            'ptychodus_input_directory': str(compute_data_posix_path),
            'ptychodus_output_directory': str(compute_data_posix_path),
            'output_data_transfer_source_endpoint': str(
                self._settings.compute_data_endpoint_id.get_value()
            ),
            'output_data_transfer_source_path': f'{compute_data_globus_path}/{StandardFileLayout.PRODUCT_OUT}',
            'output_data_transfer_destination_endpoint': str(
                self._settings.output_data_endpoint_id.get_value()
            ),
            'output_data_transfer_destination_path': f'{output_data_globus_path}/{StandardFileLayout.PRODUCT_OUT}',
            'output_data_transfer_recursive': False,
            'output_data_transfer_sync_level': self._settings.transfer_sync_level.get_value(),
        }

        flow_tags = ['aps', 'ptychography']
        job = GlobusJob(flow_input, flow_label, flow_tags)
        self._client.run_flow(job)

    def reconstruct(self, input_product_index: int, *, algorithm: str | None = None) -> None:
        self._processing_api.set_reconstructor_if_provided(algorithm)
        input_directory = self.populate_input_directory(input_product_index)
        self._run_flow('reconstruct', input_directory.name)

    def train(self, input_product_index: int, *, algorithm: str | None = None) -> None:
        # TODO customize input_directory and output_directory
        self._processing_api.set_reconstructor_if_provided(algorithm)
        input_directory = self.populate_input_directory(input_product_index)
        self._run_flow('train', input_directory.name)  # TODO mlflow

    def infer(self, input_product_index: int, *, algorithm: str | None = None) -> None:
        self._processing_api.set_reconstructor_if_provided(algorithm)
        input_directory = self.populate_input_directory(input_product_index)
        self._run_flow('infer', input_directory.name)  # TODO mlflow



from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class PtychoNNModelSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtychoNN')
        self._group.add_observer(self)

        self.num_convolution_kernels = self._group.create_integer_parameter(
            'NumberOfConvolutionKernels', 16, minimum=1
        )
        self.batch_size = self._group.create_integer_parameter('BatchSize', 64, minimum=1)
        self.use_batch_normalization = self._group.create_boolean_parameter(
            'UseBatchNormalization', False
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()


class PtychoNNTrainingSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('PtychoNNTraining')
        self._group.add_observer(self)

        self.validation_set_fractional_size = self._group.create_real_parameter(
            'ValidationSetFractionalSize', 0.1, minimum=0.0, maximum=1.0
        )
        self.max_learning_rate = self._group.create_real_parameter(
            'MaximumLearningRate', 1e-3, minimum=0.0, maximum=1.0
        )
        self.min_learning_rate = self._group.create_real_parameter(
            'MinimumLearningRate', 1e-4, minimum=0.0, maximum=1.0
        )
        self.training_epochs = self._group.create_integer_parameter('TrainingEpochs', 50, minimum=1)
        self.status_interval_in_epochs = self._group.create_integer_parameter(
            'StatusIntervalInEpochs', 1, minimum=1
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from typing import Any

from PyQt5.QtCore import Qt, QAbstractListModel, QModelIndex, QObject
from PyQt5.QtWidgets import QFormLayout, QFrame, QListView, QWidget

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.parametric import (
    BooleanParameter,
    IntegerParameter,
    RealParameter,
    StringParameter,
)

from ...model.ptychi import (
    PtyChiAffineDegreesOfFreedomBitField,
    PtyChiEnumerators,
    PtyChiProbePositionSettings,
)
from ..parametric import (
    CheckBoxParameterViewController,
    CheckableGroupBoxParameterViewController,
    ComboBoxParameterViewController,
    DecimalLineEditParameterViewController,
    DecimalSliderParameterViewController,
    IntegerLineEditParameterViewController,
    ParameterViewController,
    SpinBoxParameterViewController,
)
from .optimizer import PtyChiOptimizationPlanViewController, PtyChiOptimizerParameterViewController

__all__ = ['PtyChiProbePositionsViewController']


class PtyChiCrossCorrelationViewController(ParameterViewController, Observer):
    def __init__(
        self,
        algorithm: StringParameter,
        scale: IntegerParameter,
        real_space_width: RealParameter,
        probe_threshold: RealParameter,
    ) -> None:
        super().__init__()
        self._algorithm = algorithm
        self._scale_view_controller = SpinBoxParameterViewController(
            scale, tool_tip='Upsampling factor of the cross-correlation in real space'
        )
        self._real_space_width_view_controller = DecimalLineEditParameterViewController(
            real_space_width, tool_tip='Width of the cross-correlation in real-space'
        )
        self._probe_threshold_view_controller = DecimalSliderParameterViewController(
            probe_threshold, tool_tip='Probe intensity threshold used to calculate the probe mask'
        )
        self._widget = QFrame()
        self._widget.setFrameShape(QFrame.Shape.StyledPanel)

        layout = QFormLayout()
        layout.addRow('Scale:', self._scale_view_controller.get_widget())
        layout.addRow('Real Space Width:', self._real_space_width_view_controller.get_widget())
        layout.addRow('Probe Threshold:', self._probe_threshold_view_controller.get_widget())
        self._widget.setLayout(layout)

        algorithm.add_observer(self)
        self._sync_model_to_view()

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_model_to_view(self) -> None:
        self._widget.setVisible(self._algorithm.get_value().upper() == 'CROSS_CORRELATION')

    def _update(self, observable: Observable) -> None:
        if observable is self._algorithm:
            self._sync_model_to_view()


class PtyChiUpdateMagnitudeLimitViewController(ParameterViewController, Observer):
    def __init__(
        self,
        limit_update_magnitude: BooleanParameter,
        update_magnitude_limit: RealParameter,
    ) -> None:
        self._limit_update_magnitude = limit_update_magnitude
        self._limit_update_magnitude_view_controller = CheckBoxParameterViewController(
            limit_update_magnitude,
            'Limit Update Magnitude:',
            tool_tip='Whether to limit the update magnitude',
        )
        self._update_magnitude_limit_view_controller = DecimalLineEditParameterViewController(
            update_magnitude_limit,
            tool_tip='Maximum allowed magnitude of position update in each axis',
        )

        limit_update_magnitude.add_observer(self)
        self._sync_model_to_view()

    def get_label(self) -> QWidget:
        return self._limit_update_magnitude_view_controller.get_widget()

    def get_widget(self) -> QWidget:
        return self._update_magnitude_limit_view_controller.get_widget()

    def _sync_model_to_view(self) -> None:
        self._update_magnitude_limit_view_controller.get_widget().setEnabled(
            self._limit_update_magnitude.get_value()
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._limit_update_magnitude:
            self._sync_model_to_view()


class PtyChiSliceForCorrectionViewController(ParameterViewController, Observer):
    def __init__(
        self,
        choose_slice_for_correction: BooleanParameter,
        slice_for_correction: IntegerParameter,
    ) -> None:
        self._choose_slice_for_correction = choose_slice_for_correction
        self._choose_slice_for_correction_view_controller = CheckBoxParameterViewController(
            choose_slice_for_correction,
            'Slice For Correction:',
            tool_tip='Whether to specify the slice that is used for position correction',
        )
        self._slice_for_correction_view_controller = IntegerLineEditParameterViewController(
            slice_for_correction,
            tool_tip='Slice that is used for position correction (0-based index)',
        )

        choose_slice_for_correction.add_observer(self)
        self._sync_model_to_view()

    def get_label(self) -> QWidget:
        return self._choose_slice_for_correction_view_controller.get_widget()

    def get_widget(self) -> QWidget:
        return self._slice_for_correction_view_controller.get_widget()

    def _sync_model_to_view(self) -> None:
        self._slice_for_correction_view_controller.get_widget().setEnabled(
            self._choose_slice_for_correction.get_value()
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._choose_slice_for_correction:
            self._sync_model_to_view()


class PtyChiAffineDegreesOfFreedomListModel(QAbstractListModel):
    def __init__(self, parameter: IntegerParameter, parent: QObject | None = None) -> None:
        super().__init__(parent)
        self._dof = PtyChiAffineDegreesOfFreedomBitField(parameter)

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        value = super().flags(index)

        if index.isValid():
            value |= Qt.ItemFlag.ItemIsUserCheckable

        return value

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if index.isValid():
            if role == Qt.ItemDataRole.DisplayRole:
                return self._dof[index.row()]
            elif role == Qt.ItemDataRole.CheckStateRole:
                return (
                    Qt.CheckState.Checked
                    if self._dof.is_bit_set(index.row())
                    else Qt.CheckState.Unchecked
                )

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:  # noqa: N802
        if index.isValid() and role == Qt.ItemDataRole.CheckStateRole:
            self._dof.set_bit(index.row(), value == Qt.CheckState.Checked)
            self.dataChanged.emit(index, index)
            return True

        return False

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._dof)


class PtyChiAffineDegreesOfFreedomViewController(ParameterViewController, Observer):
    def __init__(self, parameter: IntegerParameter) -> None:
        super().__init__()
        self._parameter = parameter
        self._list_model = PtyChiAffineDegreesOfFreedomListModel(parameter)
        self._widget = QListView()
        self._widget.setModel(self._list_model)

        parameter.add_observer(self)
        self._sync_model_to_view()

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_model_to_view(self) -> None:
        self._list_model.beginResetModel()
        self._list_model.endResetModel()

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self._sync_model_to_view()


class PtyChiConstrainAffineTransformViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        is_optimizable: BooleanParameter,
        start: IntegerParameter,
        stop: IntegerParameter,
        stride: IntegerParameter,
        degrees_of_freedom: IntegerParameter,
        position_weight_update_interval: IntegerParameter,
        apply_constraint: BooleanParameter,
        max_expected_error_px: RealParameter,
        num_epochs: IntegerParameter,
    ) -> None:
        super().__init__(
            is_optimizable,
            'Constrain Affine Transform',
            tool_tip='Constrain the affine transform during position correction',
        )
        self._plan_view_controller = PtyChiOptimizationPlanViewController(
            start, stop, stride, num_epochs
        )
        self._degrees_of_freedom_view_controller = PtyChiAffineDegreesOfFreedomViewController(
            degrees_of_freedom
        )
        self._weight_update_interval_view_controller = SpinBoxParameterViewController(
            position_weight_update_interval,
            tool_tip='Interval for updating the position weight',
        )
        self._apply_constraint_view_controller = CheckBoxParameterViewController(
            apply_constraint, 'Apply Constraint', tool_tip='Whether to apply the constraint'
        )
        self._max_expected_error_px_view_controller = DecimalLineEditParameterViewController(
            max_expected_error_px, tool_tip='Maximum expected error in pixels'
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._plan_view_controller.get_widget())
        layout.addRow('Degrees of Freedom:', self._degrees_of_freedom_view_controller.get_widget())
        layout.addRow(
            'Weight Update Interval:', self._weight_update_interval_view_controller.get_widget()
        )
        layout.addRow(self._apply_constraint_view_controller.get_widget())
        layout.addRow(
            'Max Expected Error [px]:', self._max_expected_error_px_view_controller.get_widget()
        )
        self.get_widget().setLayout(layout)


class PtyChiProbePositionsViewController(CheckableGroupBoxParameterViewController):
    def __init__(
        self,
        settings: PtyChiProbePositionSettings,
        num_epochs: IntegerParameter,
        enumerators: PtyChiEnumerators,
    ) -> None:
        super().__init__(
            settings.is_optimizable,
            'Optimize Probe Positions',
            tool_tip='Whether the probe positions are optimizable',
        )
        self._optimization_plan_view_controller = PtyChiOptimizationPlanViewController(
            settings.optimization_plan_start,
            settings.optimization_plan_stop,
            settings.optimization_plan_stride,
            num_epochs,
        )
        self._optimizer_view_controller = PtyChiOptimizerParameterViewController(
            settings.optimizer, enumerators
        )
        self._step_size_view_controller = DecimalLineEditParameterViewController(
            settings.step_size, tool_tip='Optimizer step size'
        )
        self._constrain_centroid_view_controller = CheckBoxParameterViewController(
            settings.constrain_centroid,
            'Constrain Centroid',
            tool_tip='Whether to subtract the mean from positions after updating positions',
        )
        self._correction_type_view_controller = ComboBoxParameterViewController(
            settings.correction_type,
            enumerators.position_correction_types(),
            tool_tip='Algorithm used to calculate the position correction update',
        )
        self._differentiation_method_view_controller = ComboBoxParameterViewController(
            settings.differentiation_method,
            enumerators.image_gradient_methods(),
            tool_tip='Method for calculating the object gradient',
        )
        self._cross_correlation_view_controller = PtyChiCrossCorrelationViewController(
            settings.correction_type,
            settings.cross_correlation_scale,
            settings.cross_correlation_real_space_width,
            settings.cross_correlation_probe_threshold,
        )
        self._slice_for_correction_view_controller = PtyChiSliceForCorrectionViewController(
            settings.choose_slice_for_correction,
            settings.slice_for_correction,
        )
        self._clip_update_magnitude_by_mad_view_controller = CheckBoxParameterViewController(
            settings.clip_update_magnitude_by_mad,
            'Clip Update Magnitude by MAD',
            tool_tip='Whether to clip the update magnitude by the median absolute deviation',
        )
        self._update_magnitude_limit_view_controller = PtyChiUpdateMagnitudeLimitViewController(
            settings.limit_update_magnitude,
            settings.update_magnitude_limit,
        )
        self._constrain_affine_transform_view_controller = (
            PtyChiConstrainAffineTransformViewController(
                settings.constrain_affine_transform,
                settings.constrain_affine_transform_start,
                settings.constrain_affine_transform_stop,
                settings.constrain_affine_transform_stride,
                settings.constrain_affine_transform_degrees_of_freedom,
                settings.constrain_affine_transform_position_weight_update_interval,
                settings.constrain_affine_transform_apply_constraint,
                settings.constrain_affine_transform_max_expected_error_px,
                num_epochs,
            )
        )

        layout = QFormLayout()
        layout.addRow('Plan:', self._optimization_plan_view_controller.get_widget())
        layout.addRow('Optimizer:', self._optimizer_view_controller.get_widget())
        layout.addRow('Step Size:', self._step_size_view_controller.get_widget())
        layout.addRow(self._constrain_centroid_view_controller.get_widget())
        layout.addRow('Correction Type:', self._correction_type_view_controller.get_widget())
        layout.addRow(
            'Differentiation Method:', self._differentiation_method_view_controller.get_widget()
        )
        layout.addRow(self._cross_correlation_view_controller.get_widget())
        layout.addRow(
            self._slice_for_correction_view_controller.get_label(),
            self._slice_for_correction_view_controller.get_widget(),
        )
        layout.addRow(self._clip_update_magnitude_by_mad_view_controller.get_widget())
        layout.addRow(
            self._update_magnitude_limit_view_controller.get_label(),
            self._update_magnitude_limit_view_controller.get_widget(),
        )
        layout.addRow(self._constrain_affine_transform_view_controller.get_widget())
        self.get_widget().setLayout(layout)



from collections.abc import Iterator, Sequence
from importlib.metadata import version
from pathlib import Path
from typing import Final
import logging

import numpy
import ptychonn

from ptychodus.api.common import ComplexArrayType
from ptychodus.api.geometry import ImageExtent
from ptychodus.api.object import Object
from ptychodus.api.product import Product
from ptychodus.api.reconstructor import (
    LossValue,
    ReconstructInput,
    ReconstructOutput,
    TrainOutput,
    TrainableReconstructor,
)

from ..analysis import BarycentricArrayInterpolator, BarycentricArrayStitcher
from .model import PtychoNNModelProvider
from .settings import PtychoNNModelSettings, PtychoNNTrainingSettings

logger = logging.getLogger(__name__)


class CenterBoxMeanPhaseCenteringStrategy:  # TODO USE
    def __call__(self, array: ComplexArrayType) -> ComplexArrayType:
        one_third_height = array.shape[-2] // 3
        one_third_width = array.shape[-1] // 3

        amplitude = numpy.absolute(array)
        phase = numpy.angle(array)

        center_box_mean_phase = phase[
            one_third_height : one_third_height * 2, one_third_width : one_third_width * 2
        ].mean()

        return amplitude * numpy.exp(1j * (phase - center_box_mean_phase))


class PtychoNNTrainableReconstructor(TrainableReconstructor):
    MODEL_FILE_FILTER: Final[str] = 'PyTorch Lightning Checkpoint Files (*.ckpt)'
    TRAINING_DATA_FILE_FILTER: Final[str] = 'NumPy Zipped Archive (*.npz)'
    PATCHES_KEY: Final[str] = 'real'
    PATTERNS_KEY: Final[str] = 'reciprocal'

    def __init__(
        self,
        model_settings: PtychoNNModelSettings,
        training_settings: PtychoNNTrainingSettings,
        model_provider: PtychoNNModelProvider,
    ) -> None:
        self._model_settings = model_settings
        self._training_settings = training_settings
        self._model_provider = model_provider

        ptychonn_version = version('ptychonn')
        logger.info(f'\tPtychoNN {ptychonn_version}')

    @property
    def name(self) -> str:
        return self._model_provider.get_model_name()

    def get_progress_goal(self) -> int:
        return 0

    def reconstruct(self, parameters: ReconstructInput) -> Iterator[ReconstructOutput]:
        # TODO data size/shape requirements to GUI
        data = parameters.diffraction_patterns
        data_size = data.shape[-1]

        if data_size != data.shape[-2]:
            raise ValueError('PtychoNN expects square diffraction data!')

        is_data_size_pow2 = data_size & (data_size - 1) == 0 and data_size > 0

        if not is_data_size_pow2:
            raise ValueError('PtychoNN expects that the diffraction data size is a power of two!')

        model = self._model_provider.get_model()

        logger.debug('Inferring...')
        object_patches = ptychonn.infer(
            data=data.astype(numpy.float32),
            model=model,
        )

        logger.debug('Stitching...')
        object_array = parameters.product.object_.get_array()
        object_geometry = parameters.product.object_.get_geometry()
        stitcher = BarycentricArrayStitcher(
            upper=numpy.zeros_like(object_array), lower=numpy.zeros_like(object_array, dtype=float)
        )

        for scan_point, object_patch_channels in zip(
            parameters.product.probe_positions, object_patches
        ):
            patch_array = numpy.exp(1j * object_patch_channels[0])

            if object_patch_channels.shape[0] == 2:
                patch_array *= object_patch_channels[1]
            else:
                patch_array *= 0.5

            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            stitcher.add_patch(
                object_point.coordinate_x_px, object_point.coordinate_y_px, patch_array
            )

        object_ = Object(
            array=stitcher.stitch(),
            pixel_geometry=object_geometry.get_pixel_geometry(),
            center=object_geometry.get_center(),
            layer_spacing_m=parameters.product.object_.layer_spacing_m,
        )
        losses: Sequence[LossValue] = list()

        product = Product(
            metadata=parameters.product.metadata,
            probe_positions=parameters.product.probe_positions,
            probes=parameters.product.probes,
            object_=object_,
            losses=losses,
        )

        yield ReconstructOutput(product)

    def is_model_loaded(self):
        return True  # TODO

    def get_model_file_filter(self) -> str:
        return self.MODEL_FILE_FILTER

    def load_model_from_file(self, file_path: Path) -> None:
        self._model_provider.load_model_from_file(file_path)

    def save_model(self, file_path: Path) -> None:
        self._model_provider.save_model(file_path)

    def get_training_data_file_filter(self) -> str:
        return self.TRAINING_DATA_FILE_FILTER

    def export_training_data(self, file_path: Path, parameters: ReconstructInput) -> None:
        object_geometry = parameters.product.object_.get_geometry()
        interpolator = BarycentricArrayInterpolator(parameters.product.object_.get_array())
        num_channels = self._model_provider.get_num_channels()
        probe_extent = ImageExtent(
            width_px=parameters.product.probes.width_px,
            height_px=parameters.product.probes.height_px,
        )
        patches = numpy.zeros(
            (len(parameters.product.probe_positions), num_channels, *probe_extent.get_shape()),
            dtype=numpy.float32,
        )

        for index, scan_point in enumerate(parameters.product.probe_positions):
            object_point = object_geometry.map_coordinates_probe_to_object(scan_point)
            patch = interpolator.get_patch(
                object_point.coordinate_x_px,
                object_point.coordinate_y_px,
                probe_extent.width_px,
                probe_extent.height_px,
            )
            patches[index, 0, :, :] = numpy.angle(patch)

            if num_channels > 1:
                patches[index, 1, :, :] = numpy.absolute(patch)

        logger.debug(f'Writing "{file_path}" as "NPZ"')
        contents = {
            self.PATTERNS_KEY: parameters.diffraction_patterns.astype(numpy.float32),
            self.PATCHES_KEY: patches,
        }
        numpy.savez_compressed(file_path, allow_pickle=False, **contents)

    def train(self, input_path: Path, output_path: Path) -> Iterator[TrainOutput]:
        logger.debug(f'Reading "{input_path}" as "NPZ"')
        training_data = numpy.load(input_path)

        model = self._model_provider.get_model()
        logger.debug('Training...')
        training_set_fractional_size = (
            1 - self._training_settings.validation_set_fractional_size.get_value()
        )
        trainer, trainer_log = ptychonn.train(
            model=model,
            batch_size=self._model_settings.batch_size.get_value(),
            out_dir=None,
            X_train=training_data[self.PATTERNS_KEY],
            Y_train=training_data[self.PATCHES_KEY],
            epochs=self._training_settings.training_epochs.get_value(),
            training_fraction=float(training_set_fractional_size),
            log_frequency=self._training_settings.status_interval_in_epochs.get_value(),
            strategy='ddp_notebook',
        )
        self._model_provider.set_trainer(trainer)

        training_loss: list[LossValue] = []
        validation_loss: list[LossValue] = []

        for epoch, entry in enumerate(trainer_log.logs):
            try:
                tloss = LossValue(epoch, entry['training_loss'])
                vloss = LossValue(epoch, entry['validation_loss'])
            except KeyError:
                pass
            else:
                training_loss.append(tloss)
                training_loss.append(vloss)

        yield TrainOutput(
            training_loss=training_loss,
            validation_loss=validation_loss,
        )



from pathlib import Path
from typing import Final, Sequence

import numpy
import scipy.io

from ptychodus.api.geometry import PixelGeometry
from ptychodus.api.object import Object
from ptychodus.api.plugins import PluginRegistry
from ptychodus.api.probe import ProbeSequence
from ptychodus.api.product import (
    ELECTRON_VOLT_J,
    LIGHT_SPEED_M_PER_S,
    PLANCK_CONSTANT_J_PER_HZ,
    Product,
    ProductFileReader,
    ProductMetadata,
)
from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition
from ptychodus.api.reconstructor import LossValue


class FoldSliceProductFileReader(ProductFileReader):
    SIMPLE_NAME: Final[str] = 'fold_slice'
    DISPLAY_NAME: Final[str] = 'fold_slice Files (*.mat)'

    def read(self, file_path: Path) -> Product:
        point_list: list[ProbePosition] = list()

        hc_eVm = PLANCK_CONSTANT_J_PER_HZ * LIGHT_SPEED_M_PER_S / ELECTRON_VOLT_J  # noqa: N806
        mat_dict = scipy.io.loadmat(file_path, simplify_cells=True)
        p_struct = mat_dict['p']

        try:
            wavelength_m = p_struct['lambda']
        except KeyError:
            probe_energy_eV = 0.0  # noqa: N806
        else:
            probe_energy_eV = hc_eVm / wavelength_m  # noqa: N806

        try:
            tomography_angle_deg = p_struct['angle']
        except KeyError:
            tomography_angle_deg = 0.0

        metadata = ProductMetadata(
            name=file_path.stem,
            comments='',
            detector_distance_m=0.0,  # not included in file
            probe_energy_eV=probe_energy_eV,
            probe_photon_count=0.0,  # not included in file
            exposure_time_s=0.0,  # not included in file
            mass_attenuation_m2_kg=0.0,  # not included in file
            tomography_angle_deg=tomography_angle_deg,
        )

        dx_spec = p_struct['dx_spec']
        pixel_width_m = dx_spec[0]
        pixel_height_m = dx_spec[1]
        pixel_geometry = PixelGeometry(width_m=pixel_width_m, height_m=pixel_height_m)

        outputs_struct = mat_dict['outputs']
        probe_positions = outputs_struct['probe_positions']

        for idx, pos_px in enumerate(probe_positions):
            point = ProbePosition(
                idx,
                pos_px[0] * pixel_width_m,
                pos_px[1] * pixel_height_m,
            )
            point_list.append(point)

        probe_array = mat_dict['probe']

        if probe_array.ndim == 3:
            # probe_array[height, width, num_shared_modes]
            probe_array = probe_array.transpose(2, 0, 1)
        elif probe_array.ndim == 4:
            # probe_array[height, width, num_shared_modes, num_varying_modes]
            probe_array = probe_array.transpose(3, 2, 0, 1)

        probe = ProbeSequence(
            array=probe_array,
            opr_weights=None,  # TODO OPR, if available
            pixel_geometry=pixel_geometry,
        )

        object_array = mat_dict['object']

        if object_array.ndim == 3:
            # object_array[height, width, num_layers]
            object_array = object_array.transpose(2, 0, 1)

        layer_spacing_m: Sequence[float] = list()

        try:
            multi_slice_param = p_struct['multi_slice_param']
            z_distance = multi_slice_param['z_distance']
        except KeyError:
            pass
        else:
            num_spaces = object_array.shape[-3] - 1
            layer_spacing_m = numpy.squeeze(z_distance)[:num_spaces]

        object_ = Object(
            array=object_array,
            pixel_geometry=pixel_geometry,
            center=None,
            layer_spacing_m=layer_spacing_m,
        )

        fourier_error_out = outputs_struct['fourier_error_out']
        losses: list[LossValue] = []

        for epoch, value in enumerate(fourier_error_out):
            loss = LossValue(epoch, value)
            losses.append(loss)

        return Product(
            metadata=metadata,
            probe_positions=ProbePositionSequence(point_list),
            probes=probe,
            object_=object_,
            losses=losses,
        )


def register_plugins(registry: PluginRegistry) -> None:
    registry.register_product_file_reader_with_adapters(
        FoldSliceProductFileReader(),
        simple_name=FoldSliceProductFileReader.SIMPLE_NAME,
        display_name=FoldSliceProductFileReader.DISPLAY_NAME,
    )



from PyQt5.QtWidgets import QDialog, QMessageBox, QSpinBox, QWidget

from ptychodus.api.observer import Observable, Observer

from ...model.product.object import (
    DeadLeavesObjectBuilder,
    ObjectRepositoryItem,
    RandomObjectBuilder,
    STXMObjectBuilder,
)
from ..parametric import ParameterViewBuilder, ParameterViewController


class MultisliceViewController(ParameterViewController, Observer):
    def __init__(self, item: ObjectRepositoryItem) -> None:
        super().__init__()
        self._item = item
        self._parameter = item.layer_spacing_m
        self._widget = QSpinBox()

        self._sync_model_to_view()
        self._widget.valueChanged.connect(self._sync_view_to_model)
        self._parameter.add_observer(self)

    def get_widget(self) -> QWidget:
        return self._widget

    def _sync_view_to_model(self, num_layers: int) -> None:
        self._item.set_num_layers(num_layers)

    def _sync_model_to_view(self) -> None:
        self._widget.blockSignals(True)
        self._widget.setRange(1, 99)
        self._widget.setValue(self._item.get_num_layers())
        self._widget.blockSignals(False)

    def _update(self, observable: Observable) -> None:
        if observable is self._parameter:
            self._sync_model_to_view()


class ObjectEditorViewControllerFactory:
    def create_editor_dialog(
        self, item_name: str, item: ObjectRepositoryItem, parent: QWidget
    ) -> QDialog:
        object_builder = item.get_builder()
        builder_name = object_builder.get_name()
        first_layer_group = 'First Layer'
        additional_layers_group = 'Additional Layers'
        title = f'{item_name} [{builder_name}]'

        if isinstance(object_builder, DeadLeavesObjectBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_spin_box(
                object_builder.extra_padding_x, 'Extra Padding X:', group=first_layer_group
            )
            dialog_builder.add_spin_box(
                object_builder.extra_padding_y, 'Extra Padding Y:', group=first_layer_group
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.leaf_radius_lower_px,
                'Leaf Radius Lower [px]:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.leaf_radius_upper_px,
                'Leaf Radius Upper [px]:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.leaf_radius_power_law_exponent,
                'Leaf Radius Power Law Exponent:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.leaf_amplitude_lower,
                'Leaf Amplitude Lower:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.leaf_amplitude_upper,
                'Leaf Amplitude Upper:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.leaf_phase_lower_turns,
                'Leaf Phase Lower [turns]:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.leaf_phase_upper_turns,
                'Leaf Phase Upper [turns]:',
                group=first_layer_group,
            )
            dialog_builder.add_view_controller(
                MultisliceViewController(item),
                'Number of Layers:',
                group=additional_layers_group,
            )
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(object_builder, RandomObjectBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_spin_box(
                object_builder.extra_padding_x, 'Extra Padding X:', group=first_layer_group
            )
            dialog_builder.add_spin_box(
                object_builder.extra_padding_y, 'Extra Padding Y:', group=first_layer_group
            )
            dialog_builder.add_decimal_slider(
                object_builder.amplitude_mean, 'Amplitude Mean:', group=first_layer_group
            )
            dialog_builder.add_decimal_slider(
                object_builder.amplitude_deviation,
                'Amplitude Deviation:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_slider(
                object_builder.phase_deviation_turns,
                'Phase Deviation [turns]:',
                group=first_layer_group,
            )
            dialog_builder.add_decimal_line_edit(
                object_builder.blur_deviation_px, 'Blur Deviation [px]:', group=first_layer_group
            )
            dialog_builder.add_view_controller(
                MultisliceViewController(item),
                'Number of Layers:',
                group=additional_layers_group,
            )
            return dialog_builder.build_dialog(title, parent)
        elif isinstance(object_builder, STXMObjectBuilder):
            dialog_builder = ParameterViewBuilder()
            dialog_builder.add_spin_box(
                object_builder.extra_padding_x, 'Extra Padding X:', group=first_layer_group
            )
            dialog_builder.add_spin_box(
                object_builder.extra_padding_y, 'Extra Padding Y:', group=first_layer_group
            )
            dialog_builder.add_view_controller(
                MultisliceViewController(item),
                'Number of Layers:',
                group=additional_layers_group,
            )
            return dialog_builder.build_dialog(title, parent)

        return QMessageBox(
            QMessageBox.Icon.Information,
            title,
            f'"{builder_name}" has no editable parameters!',
            QMessageBox.StandardButton.Ok,
            parent,
        )



from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class FluorescenceSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Fluorescence')
        self._group.add_observer(self)

        self.file_path = self._group.create_path_parameter('FilePath', Path('/path/to/dataset.h5'))
        self.file_type = self._group.create_string_parameter('FileType', 'XRF-Maps')
        self.algorithm = self._group.create_string_parameter('Algorithm', 'VSPI')
        self.vspi_damping_factor = self._group.create_real_parameter(
            'VSPIDampingFactor', 0.0, minimum=0.0
        )
        self.vspi_max_iterations = self._group.create_integer_parameter(
            'VSPIMaxIterations', 100, minimum=1
        )
        self.upscaling_strategy = self._group.create_string_parameter('UpscalingStrategy', 'Linear')
        self.deconvolution_strategy = self._group.create_string_parameter(
            'DeconvolutionStrategy', 'Richardson-Lucy'
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from __future__ import annotations

import numpy

from ptychodus.api.probe_positions import ProbePositionSequence, ProbePosition

from .builder import ProbePositionsBuilder
from .settings import ProbePositionsSettings


class SpiralProbePositionsBuilder(ProbePositionsBuilder):
    """https://doi.org/10.1364/OE.22.012634"""

    def __init__(self, settings: ProbePositionsSettings) -> None:
        super().__init__(settings, 'spiral')
        self._settings = settings

        self.num_points = settings.num_points_x.copy()
        self.num_points.set_value(
            settings.num_points_x.get_value() * settings.num_points_y.get_value()
        )
        self._add_parameter('num_points', self.num_points)

        self._num_points = settings.num_points_y.copy()
        self._num_points.set_value(1)
        self._add_parameter('_num_points', self._num_points)

        self.radius_scalar_m = settings.radius_scalar_m.copy()
        self._add_parameter('radius_scalar_m', self.radius_scalar_m)

    def copy(self) -> SpiralProbePositionsBuilder:
        builder = SpiralProbePositionsBuilder(self._settings)

        for key, value in self.parameters().items():
            builder.parameters()[key].set_value(value.get_value())

        return builder

    def build(self) -> ProbePositionSequence:
        point_list: list[ProbePosition] = list()

        for index in range(self.num_points.get_value()):
            radius_m = self.radius_scalar_m.get_value() * numpy.sqrt(index)
            divergence_angle_rad = (3.0 - numpy.sqrt(5)) * numpy.pi
            theta_rad = divergence_angle_rad * index

            point = ProbePosition(
                index=index,
                coordinate_x_m=radius_m * numpy.cos(theta_rad),
                coordinate_y_m=radius_m * numpy.sin(theta_rad),
            )
            point_list.append(point)

        return ProbePositionSequence(point_list)



from PyQt5.QtWidgets import QWizardPage

from ptychodus.api.observer import Observable, Observer

from ....model.metadata import MetadataPresenter
from ....view.diffraction import OpenDatasetWizardMetadataPage


class OpenDatasetWizardMetadataViewController(Observer):
    def __init__(self, presenter: MetadataPresenter) -> None:
        super().__init__()
        self._presenter = presenter
        self._page = OpenDatasetWizardMetadataPage()

        presenter.add_observer(self)
        self._sync_model_to_view()
        self._page._set_complete(True)

    def import_metadata(self) -> None:
        if self._page.detector_extent_check_box.isChecked():
            self._presenter.sync_detector_extent()

        if self._page.detector_pixel_size_check_box.isChecked():
            self._presenter.sync_detector_pixel_size()

        if self._page.detector_bit_depth_check_box.isChecked():
            self._presenter.sync_detector_bit_depth()

        if self._page.detector_distance_check_box.isChecked():
            self._presenter.sync_detector_distance()

        self._presenter.sync_pattern_crop(
            sync_center=self._page.pattern_crop_center_check_box.isChecked(),
            sync_extent=self._page.pattern_crop_extent_check_box.isChecked(),
        )

        if self._page.probe_energy_check_box.isChecked():
            self._presenter.sync_probe_energy()

        if self._page.probe_photon_count_check_box.isChecked():
            self._presenter.sync_probe_photon_count()

        if self._page.exposure_time_check_box.isChecked():
            self._presenter.sync_exposure_time()

    def _sync_model_to_view(self) -> None:
        can_sync_detector_extent = self._presenter.can_sync_detector_extent()
        self._page.detector_extent_check_box.setVisible(can_sync_detector_extent)
        self._page.detector_extent_check_box.setChecked(can_sync_detector_extent)

        can_sync_detector_pixel_size = self._presenter.can_sync_detector_pixel_size()
        self._page.detector_pixel_size_check_box.setVisible(can_sync_detector_pixel_size)
        self._page.detector_pixel_size_check_box.setChecked(can_sync_detector_pixel_size)

        can_sync_detector_bit_depth = self._presenter.can_sync_detector_bit_depth()
        self._page.detector_bit_depth_check_box.setVisible(can_sync_detector_bit_depth)
        self._page.detector_bit_depth_check_box.setChecked(can_sync_detector_bit_depth)

        can_sync_detector_distance = self._presenter.can_sync_detector_distance()
        self._page.detector_distance_check_box.setVisible(can_sync_detector_distance)
        self._page.detector_distance_check_box.setChecked(can_sync_detector_distance)

        can_sync_pattern_crop_center = self._presenter.can_sync_pattern_crop_center()
        self._page.pattern_crop_center_check_box.setVisible(can_sync_pattern_crop_center)
        self._page.pattern_crop_center_check_box.setChecked(can_sync_pattern_crop_center)

        can_sync_pattern_crop_extent = self._presenter.can_sync_pattern_crop_extent()
        self._page.pattern_crop_extent_check_box.setVisible(can_sync_pattern_crop_extent)
        self._page.pattern_crop_extent_check_box.setChecked(can_sync_pattern_crop_extent)

        can_sync_probe_energy = self._presenter.can_sync_probe_energy()
        self._page.probe_energy_check_box.setVisible(can_sync_probe_energy)
        self._page.probe_energy_check_box.setChecked(can_sync_probe_energy)

        can_sync_probe_photon_count = self._presenter.can_sync_probe_photon_count()
        self._page.probe_photon_count_check_box.setVisible(can_sync_probe_photon_count)
        self._page.probe_photon_count_check_box.setChecked(can_sync_probe_photon_count)

        can_sync_exposure_time = self._presenter.can_sync_exposure_time()
        self._page.exposure_time_check_box.setVisible(can_sync_exposure_time)
        self._page.exposure_time_check_box.setChecked(can_sync_exposure_time)

    def get_widget(self) -> QWizardPage:
        return self._page

    def _update(self, observable: Observable) -> None:
        if observable is self._presenter:
            self._sync_model_to_view()



from collections.abc import Iterator
import logging


from ptychi.api import (
    AutodiffPtychographyOPRModeWeightsOptions,
    AutodiffPtychographyObjectOptions,
    AutodiffPtychographyOptions,
    AutodiffPtychographyProbeOptions,
    AutodiffPtychographyProbePositionOptions,
    AutodiffPtychographyReconstructorOptions,
    ForwardModels,
    LossFunctions,
)
from ptychi.api.task import PtychographyTask

from ptychodus.api.object import Object, ObjectGeometry
from ptychodus.api.probe import ProbeSequence
from ptychodus.api.product import LossValue, ProductMetadata
from ptychodus.api.reconstructor import ReconstructInput, ReconstructOutput, Reconstructor
from ptychodus.api.probe_positions import ProbePositionSequence

from .helper import PtyChiOptionsHelper
from .settings import PtyChiAutodiffSettings

logger = logging.getLogger(__name__)


class AutodiffReconstructor(Reconstructor):
    def __init__(
        self, options_helper: PtyChiOptionsHelper, settings: PtyChiAutodiffSettings
    ) -> None:
        super().__init__()
        self._options_helper = options_helper
        self._settings = settings
        self._epoch = 0

    @property
    def name(self) -> str:
        return 'Autodiff'

    def _create_reconstructor_options(self) -> AutodiffPtychographyReconstructorOptions:
        helper = self._options_helper.reconstructor_helper

        ####

        loss_function_str = self._settings.loss_function.get_value()

        try:
            loss_function = LossFunctions[loss_function_str.upper()]
        except KeyError:
            logger.warning('Failed to parse loss function "{loss_function_str}"!')
            loss_function = LossFunctions.MSE_SQRT

        ####

        forward_model_class_str = self._settings.forward_model_class.get_value()

        try:
            forward_model_class = ForwardModels[forward_model_class_str.upper()]
        except KeyError:
            logger.warning('Failed to parse forward model class "{forward_model_class_str}"!')
            forward_model_class = ForwardModels.PLANAR_PTYCHOGRAPHY

        ####

        return AutodiffPtychographyReconstructorOptions(
            num_epochs=helper.num_epochs,
            batch_size=helper.batch_size,
            batching_mode=helper.batching_mode,
            compact_mode_update_clustering=helper.compact_mode_update_clustering,
            compact_mode_update_clustering_stride=helper.compact_mode_update_clustering_stride,
            default_device=helper.default_device,
            default_dtype=helper.default_dtype,
            use_double_precision_for_fft=helper.use_double_precision_for_fft,
            allow_nondeterministic_algorithms=helper.allow_nondeterministic_algorithms,
            random_seed=helper.random_seed,
            displayed_loss_function=helper.displayed_loss_function,
            forward_model_options=helper.forward_model_options,
            loss_function=loss_function,
            forward_model_class=forward_model_class,
            forward_model_params=None,
        )

    def _create_object_options(self, object_: Object) -> AutodiffPtychographyObjectOptions:
        helper = self._options_helper.object_helper
        return AutodiffPtychographyObjectOptions(
            optimizable=helper.optimizable,
            optimization_plan=helper.optimization_plan,
            optimizer=helper.optimizer,
            step_size=helper.step_size,
            optimizer_params=helper.optimizer_params,
            initial_guess=helper.get_initial_guess(object_),
            slice_spacings_m=helper.get_slice_spacings_m(object_),
            slice_spacing_options=helper.slice_spacing_options,
            pixel_size_m=helper.get_pixel_size_m(object_),
            pixel_size_aspect_ratio=helper.get_pixel_aspect_ratio(object_),
            l1_norm_constraint=helper.l1_norm_constraint,
            l2_norm_constraint=helper.l2_norm_constraint,
            smoothness_constraint=helper.smoothness_constraint,
            total_variation=helper.total_variation,
            remove_grid_artifacts=helper.remove_grid_artifacts,
            multislice_regularization=helper.multislice_regularization,
            patch_interpolation_method=helper.patch_interpolation_method,
            remove_object_probe_ambiguity=helper.remove_object_probe_ambiguity,
            build_preconditioner_with_all_modes=helper.build_preconditioner_with_all_modes,
            determine_position_origin_coords_by=helper.determine_position_origin_coords_by,
            position_origin_coords=helper.get_position_origin_coords(object_),
        )

    def _create_probe_options(
        self, probes: ProbeSequence, metadata: ProductMetadata
    ) -> AutodiffPtychographyProbeOptions:
        helper = self._options_helper.probe_helper
        return AutodiffPtychographyProbeOptions(
            optimizable=helper.optimizable,
            optimization_plan=helper.optimization_plan,
            optimizer=helper.optimizer,
            step_size=helper.step_size,
            optimizer_params=helper.optimizer_params,
            initial_guess=helper.get_initial_guess(probes),
            power_constraint=helper.get_power_constraint(metadata),
            orthogonalize_incoherent_modes=helper.orthogonalize_incoherent_modes,
            orthogonalize_opr_modes=helper.orthogonalize_opr_modes,
            support_constraint=helper.support_constraint,
            center_constraint=helper.center_constraint,
            eigenmode_update_relaxation=helper.eigenmode_update_relaxation,
        )

    def _create_probe_position_options(
        self, scan: ProbePositionSequence, object_geometry: ObjectGeometry
    ) -> AutodiffPtychographyProbePositionOptions:
        helper = self._options_helper.probe_position_helper
        position_x_px, position_y_px = helper.get_positions_px(scan, object_geometry)
        return AutodiffPtychographyProbePositionOptions(
            optimizable=helper.optimizable,
            optimization_plan=helper.optimization_plan,
            optimizer=helper.optimizer,
            step_size=helper.step_size,
            optimizer_params=helper.optimizer_params,
            position_x_px=position_x_px,
            position_y_px=position_y_px,
            constrain_position_mean=helper.constrain_position_mean,
            correction_options=helper.correction_options,
            affine_transform_constraint=helper.affine_transform_constraint,
        )

    def _create_opr_mode_weight_options(
        self, probes: ProbeSequence
    ) -> AutodiffPtychographyOPRModeWeightsOptions:
        helper = self._options_helper.opr_helper
        return AutodiffPtychographyOPRModeWeightsOptions(
            optimizable=helper.optimizable,
            optimization_plan=helper.optimization_plan,
            optimizer=helper.optimizer,
            step_size=helper.step_size,
            optimizer_params=helper.optimizer_params,
            initial_weights=helper.get_initial_weights(probes),
            optimize_eigenmode_weights=helper.optimize_eigenmode_weights,
            optimize_intensity_variation=helper.optimize_intensity_variation,
            smoothing=helper.smoothing,
            update_relaxation=helper.update_relaxation,
        )

    def _create_task_options(self, parameters: ReconstructInput) -> AutodiffPtychographyOptions:
        product = parameters.product
        return AutodiffPtychographyOptions(
            data_options=self._options_helper.create_data_options(parameters),
            reconstructor_options=self._create_reconstructor_options(),
            object_options=self._create_object_options(product.object_),
            probe_options=self._create_probe_options(product.probes, product.metadata),
            probe_position_options=self._create_probe_position_options(
                product.probe_positions, product.object_.get_geometry()
            ),
            opr_mode_weight_options=self._create_opr_mode_weight_options(product.probes),
        )

    def get_progress_goal(self) -> int:
        return self._options_helper.num_epochs

    def reconstruct(self, parameters: ReconstructInput) -> Iterator[ReconstructOutput]:
        task_options = self._create_task_options(parameters)
        num_epochs = task_options.reconstructor_options.num_epochs

        task = PtychographyTask(task_options)

        with task:
            self._epoch = 0
            step_epochs = self._options_helper.num_sync_epochs

            task_reconstructor = task.reconstructor

            if task_reconstructor is None:
                raise RuntimeError('Task reconstructor is None!')

            loss_tracker = task_reconstructor.loss_tracker

            while self._epoch < num_epochs:
                task.run(step_epochs)

                losses: list[LossValue] = list()
                epoch_array = loss_tracker.table['epoch'].to_numpy()
                loss_array = loss_tracker.table['loss'].to_numpy()

                for epoch, loss in zip(epoch_array.flat, loss_array.flat):
                    loss_value = LossValue(epoch=epoch, value=loss.item())
                    losses.append(loss_value)

                product = self._options_helper.create_product(
                    product=parameters.product,
                    position_x_px=task.get_probe_positions_x(as_numpy=True),
                    position_y_px=task.get_probe_positions_y(as_numpy=True),
                    probe_array=task.get_data_to_cpu('probe', as_numpy=True),
                    object_array=task.get_data_to_cpu('object', as_numpy=True),
                    opr_weights=task.get_data_to_cpu('opr_mode_weights', as_numpy=True),
                    losses=losses,
                )

                self._epoch += step_epochs
                step_epochs = min(step_epochs, num_epochs - self._epoch)

                yield ReconstructOutput(product=product, progress=self._epoch)



from typing import Any

from PyQt5.QtCore import Qt, QAbstractTableModel, QModelIndex, QObject
from PyQt5.QtGui import QBrush

from ptychodus.api.common import BYTES_PER_MEGABYTE

from ...model.product import ProbePositionsAPI, ProbePositionsRepository
from ...model.product.probe_positions import ProbePositionsRepositoryItem


class ProbePositionsTableModel(QAbstractTableModel):
    def __init__(
        self,
        repository: ProbePositionsRepository,
        api: ProbePositionsAPI,
        editable_item_brush: QBrush,
        parent: QObject | None = None,
    ) -> None:
        super().__init__(parent)
        self._repository = repository
        self._api = api
        self._editable_item_brush = editable_item_brush
        self._header = ['Name', 'Plot', 'Builder', 'Points', 'Length [m]', 'Size [MB]']
        self._checked_item_indexes: set[int] = set()

    def insert_item(self, index: int, item: ProbePositionsRepositoryItem) -> None:
        self.beginInsertRows(QModelIndex(), index, index)
        self.endInsertRows()

    def update_item(self, index: int, item: ProbePositionsRepositoryItem) -> None:
        top_left = self.index(index, 0)
        bottom_right = self.index(index, len(self._header))
        self.dataChanged.emit(top_left, bottom_right)

    def remove_item(self, index: int, item: ProbePositionsRepositoryItem) -> None:
        self.beginRemoveRows(QModelIndex(), index, index)
        self.endRemoveRows()

    def is_item_checked(self, item_index: int) -> bool:
        return item_index in self._checked_item_indexes

    def headerData(  # noqa: N802
        self,
        section: int,
        orientation: Qt.Orientation,
        role: int = Qt.ItemDataRole.DisplayRole,
    ) -> Any:
        if orientation == Qt.Orientation.Horizontal and role == Qt.ItemDataRole.DisplayRole:
            return self._header[section]

    def data(self, index: QModelIndex, role: int = Qt.ItemDataRole.DisplayRole) -> Any:
        if not index.isValid():
            return None

        item = self._repository[index.row()]
        probe_positions = item.get_probe_positions()

        if role == Qt.ItemDataRole.DisplayRole:
            if index.column() == 0:
                return self._repository.get_name(index.row())
            elif index.column() == 1:
                return None
            elif index.column() == 2:
                return item.get_builder().get_name()
            elif index.column() == 3:
                return len(probe_positions)
            elif index.column() == 4:
                return f'{item.get_length_m():.6f}'
            elif index.column() == 5:
                return f'{probe_positions.nbytes / BYTES_PER_MEGABYTE:.2f}'
        elif role == Qt.ItemDataRole.CheckStateRole:
            if index.column() == 1:
                return (
                    Qt.CheckState.Checked
                    if index.row() in self._checked_item_indexes
                    else Qt.CheckState.Unchecked
                )
        elif role == Qt.ItemDataRole.BackgroundRole:
            if index.flags() & Qt.ItemFlag.ItemIsEditable:
                return self._editable_item_brush

    def flags(self, index: QModelIndex) -> Qt.ItemFlags:
        value = super().flags(index)

        if index.isValid():
            if index.column() in (0, 2):
                value |= Qt.ItemFlag.ItemIsEditable

            if index.column() == 1:
                value |= Qt.ItemFlag.ItemIsUserCheckable

        return value

    def setData(self, index: QModelIndex, value: Any, role: int = Qt.ItemDataRole.EditRole) -> bool:  # noqa: N802
        if not index.isValid():
            return False

        if role == Qt.ItemDataRole.EditRole:
            if index.column() == 0:
                self._repository.set_name(index.row(), str(value))
                return True
            elif index.column() == 2:
                self._api.build_probe_positions(index.row(), str(value))
                return True
        elif role == Qt.ItemDataRole.CheckStateRole:
            if index.column() == 1:
                if value == Qt.CheckState.Checked:
                    self._checked_item_indexes.add(index.row())
                else:
                    self._checked_item_indexes.discard(index.row())

                self.dataChanged.emit(index, index)

                return True

        return False

    def rowCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._repository)

    def columnCount(self, parent: QModelIndex = QModelIndex()) -> int:  # noqa: N802
        return len(self._header)



from ptychodus.api.probe import FresnelZonePlate
from ptychodus.api.plugins import PluginRegistry


def register_plugins(registry: PluginRegistry) -> None:
    registry.fresnel_zone_plates.register_plugin(
        FresnelZonePlate(160e-6, 70e-9, 60e-6),
        display_name='2-ID-D',
    )
    registry.fresnel_zone_plates.register_plugin(
        FresnelZonePlate(160e-6, 30e-9, 80e-6),
        display_name='HXN',
    )
    registry.fresnel_zone_plates.register_plugin(
        FresnelZonePlate(114.8e-6, 60e-9, 40e-6),
        display_name='LYNX',
    )
    registry.fresnel_zone_plates.register_plugin(
        FresnelZonePlate(180e-6, 15e-9, 15e-6),
        display_name='PtychoProbe',
    )
    registry.fresnel_zone_plates.register_plugin(
        FresnelZonePlate(180e-6, 50e-9, 60e-6),
        display_name='Velociprobe',
    )



from pathlib import Path

from ptychodus.api.observer import Observable, Observer
from ptychodus.api.settings import SettingsRegistry


class ObjectSettings(Observable, Observer):
    def __init__(self, registry: SettingsRegistry) -> None:
        super().__init__()
        self._group = registry.create_group('Object')
        self._group.add_observer(self)

        self.builder = self._group.create_string_parameter('Builder', 'Random')
        self.file_path = self._group.create_path_parameter('FilePath', Path('/path/to/object.npy'))
        self.file_type = self._group.create_string_parameter('FileType', 'NPY')

        self.object_layer_spacing_m = self._group.create_real_sequence_parameter(
            'ObjectLayerSpacingInMeters', []
        )

        self.extra_padding_x = self._group.create_integer_parameter('ExtraPaddingX', 1, minimum=0)
        self.extra_padding_y = self._group.create_integer_parameter('ExtraPaddingY', 1, minimum=0)
        self.amplitude_mean = self._group.create_real_parameter(
            'AmplitudeMean', 1.0, minimum=0.0, maximum=1.0
        )
        self.amplitude_deviation = self._group.create_real_parameter(
            'AmplitudeDeviation', 0.0, minimum=0.0, maximum=1.0
        )
        self.phase_deviation_turns = self._group.create_real_parameter(
            'PhaseDeviationInTurns', 0.0, minimum=0.0, maximum=1.0
        )
        self.blur_deviation_px = self._group.create_real_parameter(
            'BlurDeviation', 0.0, minimum=0.0
        )

        self.leaf_radius_lower_px = self._group.create_real_parameter(
            'LeafRadiusLowerInPixels', 2.0, minimum=0.0
        )
        self.leaf_radius_upper_px = self._group.create_real_parameter(
            'LeafRadiusUpperInPixels', 999.0, minimum=0.0
        )
        self.leaf_radius_power_law_exponent = self._group.create_real_parameter(
            'LeafRadiusPowerLawExponent', 3.0, minimum=1.0
        )
        self.leaf_amplitude_lower = self._group.create_real_parameter(
            'LeafAmplitudeLower', 0.0, minimum=0.0, maximum=1.0
        )
        self.leaf_amplitude_upper = self._group.create_real_parameter(
            'LeafAmplitudeUpper', 1.0, minimum=0.0, maximum=1.0
        )
        self.leaf_phase_lower_turns = self._group.create_real_parameter(
            'LeafPhaseLowerInTurns',
            -0.5,
        )
        self.leaf_phase_upper_turns = self._group.create_real_parameter(
            'LeafPhaseUpperInTurns',
            0.5,
        )

    def _update(self, observable: Observable) -> None:
        if observable is self._group:
            self.notify_observers()



from __future__ import annotations
from decimal import Decimal

from PyQt5.QtCore import pyqtSignal
from PyQt5.QtWidgets import QComboBox, QHBoxLayout, QWidget
import numpy

from .decimal_line_edit import DecimalLineEdit


class AngleWidget(QWidget):
    angle_changed = pyqtSignal(Decimal)

    def __init__(self, parent: QWidget | None = None) -> None:
        super().__init__(parent)
        self.angle_in_turns = Decimal()
        self.angle_line_edit = DecimalLineEdit.create_instance(is_signed=False)
        self.units_combo_box = QComboBox()

        self.angle_line_edit.value_changed.connect(self._set_angle_in_turns_from_widgets)

        self.units_combo_box.addItem('turn', Decimal(1))
        self.units_combo_box.addItem('deg', Decimal(360))
        self.units_combo_box.addItem('rad', 2 * Decimal.from_float(numpy.pi))
        self.units_combo_box.activated.connect(self._update_display)

        layout = QHBoxLayout()
        layout.setContentsMargins(0, 0, 0, 0)
        layout.addWidget(self.angle_line_edit)
        layout.addWidget(self.units_combo_box)
        self.setLayout(layout)

    def is_read_only(self) -> bool:
        return self.angle_line_edit.is_read_only()

    def set_read_only(self, enable: bool) -> None:
        self.angle_line_edit.set_read_only(enable)

    def get_angle_in_turns(self) -> Decimal:
        return self.angle_in_turns

    def set_angle_in_turns(self, angle_in_turns: Decimal) -> None:
        self.angle_in_turns = angle_in_turns
        self._update_display()
        self.angle_changed.emit(self.get_angle_in_turns())

    def _set_angle_in_turns_from_widgets(self, angle: Decimal) -> None:
        self.angle_in_turns = angle / self.units_combo_box.currentData()
        self.angle_changed.emit(self.angle_in_turns)

    def _update_display(self) -> None:
        angle_in_display_units = self.angle_in_turns * self.units_combo_box.currentData()
        self.angle_line_edit.set_value(angle_in_display_units)
